#include "ObjectRelationalDataBuffer.h"
#include "ContainerHandlerFactory.h"
#include "IContainerHandler.h"
#include "RelationalStorageServiceException.h"
#include "DataElement.h"

#include "POOLCore/IBlobStreamingService.h"
#include "POOLCore/Context.h"

#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"

#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"

#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Blob.h"

#include "Reflex/Member.h"
#include "Reflex/Base.h"


bool pool::RelationalStorageService::ObjectRelationalDataBuffer::existAttribute(const std::string& attributeName, const coral::AttributeList& data){
  bool found = false;
  for(coral::AttributeList::const_iterator iAttr = data.begin();
      iAttr!=data.end() && !found; ++iAttr){
    if( iAttr->specification().name() == attributeName ) found = true;
  }
  return found;
}

pool::RelationalStorageService::ObjectRelationalDataBuffer::ObjectRelationalDataBuffer( coral::ISchema& schema, pool::Context* context):
  m_schema( schema ),
  m_context(context),
  m_mapping( 0 ),
  m_mainDataElement(0),
  m_dataElements(),
  m_parentDataElement(0),
  m_metadata(),
  m_parentMetadata( 0 ),
  m_parentMetadataOffset( 0 ),
  m_data( 0 ),
  m_table( 0 ),
  m_dictionary(),
  m_keyItems( 0 ),
  m_keyData( false ),
  m_containerHandler( 0 ),
  m_arrayBuffers(),
  m_blobWriters(),
  m_poolContainerBuffers(),
  m_poolPointerBuffers(),
  m_poolPolymorphicPointerBuffers(),
  m_dependencies(),
  m_updateClause(""),
  m_idClause("")
{
  m_mainDataElement = new DataElement;
}

pool::RelationalStorageService::ObjectRelationalDataBuffer::~ObjectRelationalDataBuffer()
{
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_arrayBuffers.begin();
        iBuffer != m_arrayBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_arrayBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_poolContainerBuffers.begin();
        iBuffer != m_poolContainerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolContainerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_poolPointerBuffers.begin();
        iBuffer != m_poolPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_poolPolymorphicPointerBuffers.begin();
        iBuffer != m_poolPolymorphicPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPolymorphicPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_dependencies.begin();
        iBuffer != m_dependencies.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_dependencies.clear();
  
  if(m_mainDataElement) delete m_mainDataElement;
  m_dataElements.clear();
  
  if ( m_containerHandler ) delete m_containerHandler;
  if ( m_data ) delete m_data;
  for( std::map< std::string,pool::IBlobWriter* >::iterator iBlobWriter=m_blobWriters.begin();
       iBlobWriter!=m_blobWriters.end();
       iBlobWriter++)
  {
    delete iBlobWriter->second;
  } 
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::build( const Reflex::Type& dict,
                                                                   const pool::ObjectRelationalMappingElement& mapping )
{
  m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  m_mapping = &mapping;
  // resolve possible typedef chains
  Reflex::Type dictionary = ObjectRelationalClassUtils::resolvedType(dict);
  m_dictionary = dictionary;
  m_data = new coral::AttributeList;
  this->buildBuffer( dictionary, mapping, *m_mainDataElement, false );
  // Build the actual buffer
  return true;
}

pool::RelationalStorageService::ObjectRelationalDataBuffer&
pool::RelationalStorageService::ObjectRelationalDataBuffer::addDependency()
{
  ObjectRelationalDataBuffer* newBuffer = new ObjectRelationalDataBuffer(m_schema, m_context);
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = m_mainDataElement;
  m_dependencies.push_back( newBuffer );
  return *newBuffer;
}

int
pool::RelationalStorageService::ObjectRelationalDataBuffer::processBaseDataMembers( const Reflex::Type& dictionary,
                                                                                    const pool::ObjectRelationalMappingElement& mapping,
                                                                                    DataElement& dataElement,
                                                                                    bool keyData ){
  int retVal = 0;
  for ( unsigned int i=0;i<dictionary.BaseSize();i++){
    Reflex::Base base = dictionary.BaseAt(i);
    Reflex::Type baseType = ObjectRelationalClassUtils::resolvedType( base.ToType() );
    retVal += processBaseDataMembers( baseType, mapping, dataElement, keyData );
    for ( unsigned int j=0;j<baseType.DataMemberSize();j++){
      Reflex::Member dataMember = baseType.DataMemberAt(j);      
      // Ignore the transients and the statics
      if ( dataMember.IsTransient() || dataMember.IsStatic() ) continue;
      // Get the member type and resolve possible typedef chains
      Reflex::Type dataMemberType = ObjectRelationalClassUtils::resolvedType( dataMember.TypeOf() );
      if ( ! dataMemberType ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for data member \"" +
                                                                                 dataMember.Name() + "\" of class \"" +
                                                                                 baseType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectRelationalDataBuffer::processBaseDataMembers" );
      }
      
      // check if the member is from a class in the inheritance tree
      Reflex::Type declaringType = ObjectRelationalClassUtils::resolvedType( dataMember.DeclaringType());
      std::string scope = declaringType.Name(Reflex::SCOPED|Reflex::FINAL);
      // Get the data member name
      std::string dataMemberName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( dataMember.Name(), scope );

      // Retrieve the relevant mapping element
      pool::ObjectRelationalMappingElement::const_iterator iDataMemberMapping = mapping.find( dataMemberName );
      if ( iDataMemberMapping == mapping.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Data member \"" + dataMemberName + "\" not found in the mapping element",
                                                                                 "ObjectRelationalDataBuffer::processBaseDataMembers" );
      }
      const pool::ObjectRelationalMappingElement& dataMemberMapping = iDataMemberMapping->second;
      DataElement& dataMemberElement = dataElement.addChild( dataMember.Offset(), base.OffsetFP() );
      retVal += this->buildBuffer( dataMemberType, dataMemberMapping, dataMemberElement, keyData );
    }
  }
  return retVal;
}

int
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildBuffer( const Reflex::Type& dictionary,
                                                                         const pool::ObjectRelationalMappingElement& mapping,
                                                                         pool::RelationalStorageService::DataElement& dataElement,
                                                                         bool keyData )
{
  int retVal = 0;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Primitive ) { // Primitives
    if ( !( dictionary.IsFundamental() || dictionary.TypeInfo() == typeid(std::string) || dictionary.IsEnum()) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               "\" is not declared as primitive in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    if(!existAttribute( mapping.columnNames()[0], *m_data )){
      const std::type_info* attrType = &dictionary.TypeInfo();
      if(dictionary.IsEnum()) attrType = &typeid(int);
      m_data->extend( mapping.columnNames()[0], *attrType );
    }
    m_dataElements.push_back( &dataElement );
    ++retVal;
    appendToUpdateClause( mapping.columnNames()[0] );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Blob ) { // Blobs
    if(!existAttribute( mapping.columnNames()[0], *m_data )){
      m_data->extend<coral::Blob>( mapping.columnNames()[0] );
    }
    m_dataElements.push_back( &dataElement );
    ++retVal;
    appendToUpdateClause( mapping.columnNames()[0] );
    // book also the corresponding Blob Writer
    pool::IBlobWriter* blobWriter = 0;
    std::map< std::string,pool::IBlobWriter* >::const_iterator iBlobWriter=m_blobWriters.find(mapping.columnNames()[0]);
    if( iBlobWriter!=m_blobWriters.end() )
    {
      blobWriter = iBlobWriter->second;
    }
    if(!blobWriter)
    {
      pool::IBlobStreamingService* blobStreamingService = m_context->blobStreamer();
      if ( !blobStreamingService ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot stream to blob variable \""+mapping.variableName()+"\": No blob streaming Service registered.","ObjectRelationalDataBuffer::buildBuffer" );
      }
      // fix me : version is currently not used...
      blobWriter = blobStreamingService->newWriter(Reflex::Type::ByName(mapping.variableType()),"");
      m_blobWriters.insert(std::make_pair(mapping.columnNames()[0],blobWriter));
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Object ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::DependentObject ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) { // Objects
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
      columnNames.pop_back();
      columnNames.pop_back();
    }
    bool setIdClause = false;
    if(m_idClause.empty()) setIdClause = true;
    unsigned int iCol=0;
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn, ++iCol ) {
      if(!existAttribute( *iColumn, *m_data )){
        const std::type_info* colType = &typeid(int);
        m_data->extend( *iColumn, *colType );
      }
      if(setIdClause) {
        if(!m_idClause.empty()) m_idClause.append(" AND ");
        m_idClause.append( *iColumn+" = :"+*iColumn );
      }
    }

    if ( !ObjectRelationalClassUtils::isTypeObject( dictionary ) ) {
      // processing of the non-objects (top level: primitives, containers, arrays, ptr's, etc)
      ObjectRelationalMappingElement::const_iterator iElement = mapping.find( dictionary.Name(Reflex::SCOPED) );
      if(iElement==mapping.end()) throw pool::RelationalStorageService::RelationalStorageServiceException( "Mapping for type \"" + dictionary.Name(Reflex::SCOPED) +
                                                                                                           "\" has not been found.",
                                                                                                           "ObjectRelationalDataBuffer::buildBuffer" );
      retVal += this->buildBuffer( dictionary, iElement->second, dataElement, keyData );
    } else {
      if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {

        dataElement.setRefFlag();
        // Book the OID columns
        size_t index = columnNames.size();
        m_data->extend< int >( mapping.columnNames()[index] );
        m_dataElements.push_back( &dataElement );
        appendToUpdateClause(mapping.columnNames()[index]);
        ++index;
        m_data->extend< int >( mapping.columnNames()[index] );
        m_dataElements.push_back( &dataElement );
        appendToUpdateClause(mapping.columnNames()[index]); 
      } else {

        // processing base classes data members
        retVal += processBaseDataMembers( dictionary, mapping, dataElement, keyData );
        
        // Loop over the data members of the class.
        for ( unsigned int i=0;i<dictionary.DataMemberSize();i++){

          Reflex::Member dataMember = dictionary.DataMemberAt(i);
          
          Reflex::Type declaringType = ObjectRelationalClassUtils::resolvedType( dataMember.DeclaringType());
          if( declaringType != dictionary ){
            continue;
          }
          
          // Ignore the transients and the statics
          if ( dataMember.IsTransient() || dataMember.IsStatic() ) continue;         

          // Get the member type and resolve possible typedef chains
          Reflex::Type dataMemberType = ObjectRelationalClassUtils::resolvedType( dataMember.TypeOf() );
          if ( ! dataMemberType ) {
            throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for data member \"" +
                                                                                     dataMember.Name() + "\" of class \"" +
                                                                                     dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                     "ObjectRelationalDataBuffer::buildBuffer" );
          }
      
          // check if the member is from a class in the inheritance tree
          std::string scope("");
          // Get the data member name
          std::string dataMemberName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( dataMember.Name(), scope );

          // Retrieve the relevant mapping element
          pool::ObjectRelationalMappingElement::const_iterator idataMemberMapping = mapping.find( dataMemberName );
          if ( idataMemberMapping == mapping.end() ) {
            throw pool::RelationalStorageService::RelationalStorageServiceException( "Data member \"" + dataMemberName + "\" not found in the mapping element",
                                                                                     "ObjectRelationalDataBuffer::buildBuffer" );
          }
          const pool::ObjectRelationalMappingElement& dataMemberMapping = idataMemberMapping->second;
          DataElement& dataMemberElement = dataElement.addChild( dataMember.Offset(), 0 );

          retVal += this->buildBuffer( dataMemberType, dataMemberMapping, dataMemberElement, keyData );
        }
      }
      
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::InlineCArray ) { // InlineCArrays
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn ) {
      if(!existAttribute( *iColumn, *m_data )){
        m_data->extend< int >( *iColumn );
      }
      appendToUpdateClause(*iColumn);
    }

      // Get the member type and resolve possible typedef chains
    Reflex::Type arrayType = ObjectRelationalClassUtils::resolvedType( dictionary.ToType() );
    if ( ! arrayType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException(
        "Missing dictionary information for the element of array \"" +dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"","ObjectRelationalDataBuffer::buildBuffer" );
    }
    // Loop over the elements of the array.
    for ( unsigned int i=0;i<dictionary.ArrayLength();i++){

      // Form the element name
      std::string arrayElementLabel = ObjectRelationalDefaultMappingRules::variableNameForArrayIndex( mapping.variableName(),i);

      // Retrieve the relevant mapping element
      pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( arrayElementLabel );
      if ( iMappingElement == mapping.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Array Mapping Element \"" + arrayElementLabel + "\" not found in the mapping element",
                                                                                 "ObjectRelationalDataBuffer::buildBuffer" );
      }
      const pool::ObjectRelationalMappingElement& mappingElement = iMappingElement->second;
      DataElement& arrayElement = dataElement.addChild( i*arrayType.SizeOf(), 0 );
      
      retVal += this->buildBuffer( arrayType, mappingElement, arrayElement, keyData );
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Array ) { // Containers
    if ( ! ObjectRelationalClassUtils::isTypeContainer( dictionary ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as Array in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildArrayBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolArray) { // PoolArrays
    if ( ! ObjectRelationalClassUtils::isTypePVector( dictionary ) && ! ObjectRelationalClassUtils::isTypePContainer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as PoolArray in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    if( ObjectRelationalClassUtils::isTypePVector( dictionary ) )
      this->buildArrayBuffer( dictionary, mapping, dataElement, keyData );
    else if ( ObjectRelationalClassUtils::isTypePContainer( dictionary ))
      this->buildPoolContainerBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::CArray ) { // C-Arrays
    if ( ! dictionary.IsArray() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a C-Array in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildCArrayBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPointer ) { // Pool Pointers
    if ( ! ObjectRelationalClassUtils::isTypePoolPointer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a Pool Pointer in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildPoolPointerBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPolymorphicPointer ) { // Pool Poly Pointers
    if ( ! ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a Pool PolymorphicPointer in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    if(!existAttribute( mapping.columnNames()[0], *m_data )){
      m_data->extend( mapping.columnNames()[0], typeid(std::string) );
    }
    if(!existAttribute( mapping.columnNames()[1], *m_data )){
      m_data->extend( mapping.columnNames()[1], typeid(int) );
    }
    int metadataOffset = m_metadata.size();
    dataElement.setMetadata( metadataOffset );
    m_dataElements.push_back( &dataElement );    
    m_metadata.push_back( std::make_pair(mapping.variableType(),0) );
    ++retVal;
    appendToUpdateClause( mapping.columnNames()[0] );
    appendToUpdateClause( mapping.columnNames()[1] );
    this->buildPoolPolymorphicPointerBuffer( dictionary, *m_mapping, mapping, dataElement, metadataOffset, keyData );
  }
  else { // Do not treat other types yet...
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot treat anything else than objects, primitives, C-arrays and containers for the moment...",
                                                                             "ObjectRelationalDataBuffer::buildBuffer" );
  }

  return retVal;
}




bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildArrayBuffer( const Reflex::Type& dictionary,
                                                                              const pool::ObjectRelationalMappingElement& mapping,
                                                                              pool::RelationalStorageService::DataElement& dataElement,
                                                                              bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_schema,
                                                                                                                                         m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );
  if ( ! newBuffer->m_containerHandler ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not create a container handler",
                                                                             "ObjectRelationalDataBuffer::buildArrayBuffer" );
  }

  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  size_t i=0;
  if( mapping.isDependentTree() ){
      newBuffer->m_data->extend< int >( columnNames[i] );
      i++;
  }
  for ( ; i<columnNames.size(); ++i ){
    newBuffer->m_data->extend< int >( columnNames[i] );
  }

  Reflex::Type valueType;
  // Get the type of the container...
  if ( ObjectRelationalClassUtils::isTypeAssociativeContainer(dictionary)){
    
    Reflex::Type keyType = ObjectRelationalClassUtils::containerKeyType(dictionary);

    Reflex::Type keyResolvedType = ObjectRelationalClassUtils::resolvedType(keyType);
    if ( ! keyType || !keyResolvedType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the key type of the container \"" +
                                                                               dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                               "ObjectRelationalDataBuffer::buildArrayBuffer" );
    }
    std::string keyName = keyType.Name();
    // Retrieve the relevant mapping element
    pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( keyName );
    if ( iMappingElement == mapping.end() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + keyName + "\" not found in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildArrayBuffer" );
    }
    newBuffer->m_keyItems = newBuffer->buildBuffer( keyResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, true );
    valueType = ObjectRelationalClassUtils::containerDataType(dictionary);
  } else {
    valueType = ObjectRelationalClassUtils::containerValueType(dictionary);
  }

  Reflex::Type valueResolvedType = ObjectRelationalClassUtils::resolvedType(valueType);
  // Check the component type
  if ( ! valueType || !valueResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the content type of the container \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildArrayBuffer" );
  }
  
  std::string valueName = valueType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( valueName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + valueName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildArrayBuffer" );
  }
  newBuffer->buildBuffer( valueResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildCArrayBuffer( const Reflex::Type& dictionary,
                                                                               const pool::ObjectRelationalMappingElement& mapping,
                                                                               pool::RelationalStorageService::DataElement& dataElement,
                                                                               bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_schema,
                                                                                                                                         m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );
  if ( ! newBuffer->m_containerHandler ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not create a container handler",
                                                                             "ObjectRelationalDataBuffer::buildCArrayBuffer" );
  }

  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
      newBuffer->m_data->extend< int >( *iColumn );
      iColumn++;
  }
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_data->extend< int >( *iColumn );
  }
  
  // Check the array type
  Reflex::Type arrayType = dictionary.ToType();
  Reflex::Type arrayResolvedType = ObjectRelationalClassUtils::resolvedType(arrayType);
  // Check the component type
  if ( ! arrayType || !arrayResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the element type of the array \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildCArrayBuffer" );
  }
  
  std::string arrayTypeName = arrayType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( arrayTypeName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + arrayTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildCArrayBuffer" );
  }
  newBuffer->buildBuffer( arrayResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildPoolContainerBuffer( const Reflex::Type& dictionary,
                                                                                      const pool::ObjectRelationalMappingElement& mapping,
                                                                                      DataElement& dataElement,
                                                                                      bool keyData ){
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_schema,
                                                                                                                                         m_context);
  m_poolContainerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  Reflex::Type storeBaseType = ObjectRelationalClassUtils::containerSubType(dictionary,"store_base_type");
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( storeBaseType );
  if ( ! newBuffer->m_containerHandler ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not create a container handler",
                                                                             "ObjectRelationalDataBuffer::buildPoolContainerBuffer" );
  }

  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  size_t i=0;
  if( mapping.isDependentTree() ){
      newBuffer->m_data->extend< int >( columnNames[i] );
      i++;
  }
  for ( ; i<columnNames.size(); ++i ){
    newBuffer->m_data->extend< int >( columnNames[i] );
  }

  // here the true type (not the pair<size_t,T>) has to be considered.
  Reflex::Type valueType = ObjectRelationalClassUtils::containerValueType(dictionary);
  Reflex::Type valueResolvedType = ObjectRelationalClassUtils::resolvedType(valueType);
  // Check the component type
  if ( ! valueType || !valueResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the content type of the container \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildPoolContainerBuffer" );
  }
  
  std::string valueName = valueType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( valueName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + valueName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildPoolContainerBuffer" );
  }
  newBuffer->buildBuffer( valueResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildPoolPointerBuffer( const Reflex::Type& dictionary,
                                                                                    const pool::ObjectRelationalMappingElement& mapping,
                                                                                    pool::RelationalStorageService::DataElement& dataElement,
                                                                                    bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_schema,
                                                                                                                                         m_context);
  m_poolPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
      newBuffer->m_data->extend< int >( *iColumn );
      iColumn++;
  }
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_data->extend< int >( *iColumn );
  }

  // Check the  type
  Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildPoolPointerBuffer" );
  }
  
  std::string ptrTypeName = ptrType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( ptrTypeName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + ptrTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildPoolPointerBuffer" );
  }
  newBuffer->buildBuffer( ptrResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildPoolPolymorphicPointerBuffer( const Reflex::Type& dictionary,
                                                                                               const pool::ObjectRelationalMappingElement& topLevelMapping,
                                                                                               const pool::ObjectRelationalMappingElement& polyPtrMapping,
                                                                                               
                                                                                               pool::RelationalStorageService::DataElement& dataElement,
                                                                                               int metadataOffset,
                                                                                               bool keyData)
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_schema,
                                                                                                                                         m_context);
  m_poolPolymorphicPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &topLevelMapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_parentMetadata = &m_metadata;
  newBuffer->m_parentMetadataOffset = metadataOffset;
  newBuffer->m_keyData = keyData;

  // only for the update...
  newBuffer->m_data = new coral::AttributeList;
  const std::vector< std::string >& columnNames = polyPtrMapping.columnNames();
  newBuffer->m_data->extend< std::string >( columnNames[0] );
  newBuffer->m_data->extend< int >( columnNames[1] );

  // Check the  type
  Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildPoolPolymorphicPointerBuffer" );
  }
  return true;
}

void
pool::RelationalStorageService::ObjectRelationalDataBuffer::appendToUpdateClause(const std::string& columnName){
  std::string columnSetClause(columnName+" = :"+columnName);
  if(!m_updateClause.empty()) m_updateClause.append(", ");
  m_updateClause.append(columnSetClause);
}

coral::AttributeList&
pool::RelationalStorageService::ObjectRelationalDataBuffer::data()
{
  return *m_data;
}

const std::vector<pool::RelationalStorageService::DataElement*>&
pool::RelationalStorageService::ObjectRelationalDataBuffer::dataElements() const
{
  return m_dataElements;
}

pool::RelationalStorageService::DataElement&
pool::RelationalStorageService::ObjectRelationalDataBuffer::parentDataElement() const
{
  if(!m_parentDataElement) throw RelationalStorageServiceException("Top level buffer does not have parent DataElement.","ObjectRelationalDataBuffer::parentDataElement");
  return *m_parentDataElement;
}


const std::vector< std::pair<std::string,int> >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::metadata() const {
  return m_metadata;
}

std::vector< std::pair<std::string,int>  >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::metadata() {
  return m_metadata;
}

const std::vector< std::pair<std::string,int> >*
pool::RelationalStorageService::ObjectRelationalDataBuffer::parentMetadata() const{
  return m_parentMetadata;
}

int
pool::RelationalStorageService::ObjectRelationalDataBuffer::parentMetadataOffset() const{
  return m_parentMetadataOffset;
}

int
pool::RelationalStorageService::ObjectRelationalDataBuffer::keyItems() const
{
  return m_keyItems;
}

coral::ITable&
pool::RelationalStorageService::ObjectRelationalDataBuffer::table()
{
  return *m_table;
}

const Reflex::Type&
pool::RelationalStorageService::ObjectRelationalDataBuffer::dictionary() const
{
  return m_dictionary;
}

pool::RelationalStorageService::IContainerHandler&
pool::RelationalStorageService::ObjectRelationalDataBuffer::containerHandler()
{
  return *m_containerHandler;
}


bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::isKeyData() const
{
  return m_keyData;
}


pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalDataBuffer::begin()
{
  return m_arrayBuffers.begin();
}

pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalDataBuffer::end()
{
  return m_arrayBuffers.end();
}

size_t
pool::RelationalStorageService::ObjectRelationalDataBuffer::size() const
{
  return m_arrayBuffers.size()+m_poolContainerBuffers.size()+m_poolPointerBuffers.size()+m_poolPolymorphicPointerBuffers.size();
}

const std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::poolContainerBuffers() const {
  return m_poolContainerBuffers;
}

const std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::poolPointerBuffers() const {
  return m_poolPointerBuffers;
}

const std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::poolPolymorphicPointerBuffers() const {
  return m_poolPolymorphicPointerBuffers;
}

const pool::ObjectRelationalMappingElement& pool::RelationalStorageService::ObjectRelationalDataBuffer::mapping() const {
  return *m_mapping;
}

/// Returns a blob version of the specified data, using the blob streamer registered in the pool context
/// The offset is used to lookup the writer in the map
const coral::Blob&
pool::RelationalStorageService::ObjectRelationalDataBuffer::writeBlob(const std::string& columnName, const void* data)
{
  pool::IBlobWriter* blobWriter = 0;
  
  std::map< std::string,pool::IBlobWriter* >::const_iterator iBlobWriter=m_blobWriters.find(columnName);
  if( iBlobWriter!=m_blobWriters.end() )
  {
    blobWriter = iBlobWriter->second;
  } else 
  {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot stream to blob: The streamer has not been booked.","ObjectRelationalDataBuffer::writeBlob" );
  }
  return blobWriter->write(data);
}

const std::string& 
pool::RelationalStorageService::ObjectRelationalDataBuffer::updateClause() const{
  return m_updateClause;
}

const std::string& 
pool::RelationalStorageService::ObjectRelationalDataBuffer::idClause() const{
  return m_idClause;
}
