#ifndef OBJECTRELATIONALDATABUFFER_H
#define OBJECTRELATIONALDATABUFFER_H

#include <vector>
#include <map>
#include "Reflex/Type.h"
#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

namespace coral {

  class ITable;
  class ISchema;
  class AttributeListSpecification;
  class AttributeList;
  class Blob;
}

namespace pool {

  class ObjectRelationalMappingElement;
  class IBlobWriter;
  class Context;

  namespace RelationalStorageService {

    class IContainerHandler;
    class DataElement;
    
    class ObjectRelationalDataBuffer {
    public:
      /// Constuctor
      ObjectRelationalDataBuffer( coral::ISchema& schema, pool::Context* context );

      /// Destructor
      ~ObjectRelationalDataBuffer();

      /// Builds the structure given a mapping element and the dictionary information
      bool build( const Reflex::Type& dictionary ,
                  const pool::ObjectRelationalMappingElement& mapping );

      /// adds a new buffer to the dependency list
      ObjectRelationalDataBuffer& addDependency();

      /// Returns the data buffer
      coral::AttributeList& data();

      /// Returns the children data elements
      const std::vector<DataElement*>& dataElements() const;

      /// Returns the parent data elements
      DataElement& parentDataElement() const;

      /// Returns the metadata vector
      const std::vector< std::pair<std::string,int> >& metadata() const;
      
      /// Returns the metadata vector
      std::vector< std::pair<std::string,int> >& metadata();

      /// Returns the parent metadata
      const std::vector< std::pair<std::string,int> >* parentMetadata() const;
      
      /// Returns the parent metadata offset
      int parentMetadataOffset() const;

      /// Returns the number of items belonging to the key (in case of an associative container)
      int keyItems() const;

      /// Returns the associoted table handle
      coral::ITable& table();

      /// Returns the associated dictionary information
      const Reflex::Type& dictionary() const;

      /// Returns the container handler
      IContainerHandler& containerHandler();

      /// Returns a flag indicating whether the starting address is the key of an associative container
      bool isKeyData() const;

      /// Returns a blob version of the specified data, using the blob streamer registered in the pool context 
      const coral::Blob& writeBlob(const std::string& columnName, const void* data);

      const std::string& updateClause() const;

      const std::string& idClause() const;

    public:
      typedef std::vector< ObjectRelationalDataBuffer* >::iterator iterator;
      iterator begin();
      iterator end();
      size_t size() const;
      const std::vector< ObjectRelationalDataBuffer* >& poolContainerBuffers() const;
      const std::vector< ObjectRelationalDataBuffer* >& poolPointerBuffers() const;
      const std::vector< ObjectRelationalDataBuffer* >& poolPolymorphicPointerBuffers() const;

      const ObjectRelationalMappingElement& mapping() const;

      private:

      /// process the data members from the base trees 
      int processBaseDataMembers( const Reflex::Type& dictionary,
                                  const pool::ObjectRelationalMappingElement& mapping,
                                  DataElement& dataElement,
                                  bool keyData );
      
      /// builds a buffer entry
      int buildBuffer( const Reflex::Type& dictionary,
                       const pool::ObjectRelationalMappingElement& mapping,
                       DataElement& dataElement,
                       bool keyData );

      /// builds an array buffer
      bool buildArrayBuffer( const Reflex::Type& dictionary,
                             const pool::ObjectRelationalMappingElement& mapping,
                             DataElement& dataElement,
                             bool keyData );
      
      /// builds a C-array buffer
      bool buildCArrayBuffer( const Reflex::Type& dictionary,
                              const pool::ObjectRelationalMappingElement& mapping,
                              DataElement& dataElement,
                              bool keyData );
      
      /// builds a Pool Container buffer
      bool buildPoolContainerBuffer( const Reflex::Type& dictionary,
                                     const pool::ObjectRelationalMappingElement& mapping,
                                     DataElement& dataElement,
                                     bool keyData );
      /// builds a Pool Ptr buffer
      bool buildPoolPointerBuffer( const Reflex::Type& dictionary,
                                   const pool::ObjectRelationalMappingElement& mapping,
                                   DataElement& dataElement,
                                   bool keyData );

      /// builds a Pool PolyPtr buffer
      bool buildPoolPolymorphicPointerBuffer( const Reflex::Type& dictionary,
                                              const pool::ObjectRelationalMappingElement& topLevelMapping,
                                              const pool::ObjectRelationalMappingElement& polyPtrMapping,
                                              DataElement& dataElement,
                                              int metadataOffset,
                                              bool keyData);

      /// add the column to the buffered update clause
      void appendToUpdateClause(const std::string& columnName);

      static bool existAttribute(const std::string& attributeName, const coral::AttributeList& data);

      private:

      coral::ISchema& m_schema;
      pool::Context* m_context;
      const pool::ObjectRelationalMappingElement* m_mapping;
      DataElement* m_mainDataElement;
      std::vector<DataElement*> m_dataElements;
      DataElement* m_parentDataElement;
      std::vector< std::pair<std::string,int> >   m_metadata;
      std::vector< std::pair<std::string,int> >*  m_parentMetadata;
      int m_parentMetadataOffset;
      coral::AttributeList* m_data;
      coral::ITable* m_table;
      Reflex::Type m_dictionary;
      int m_keyItems;
      bool m_keyData;
      IContainerHandler* m_containerHandler;
      std::vector< ObjectRelationalDataBuffer* > m_arrayBuffers;
      std::map< std::string, pool::IBlobWriter* > m_blobWriters;
      std::vector< ObjectRelationalDataBuffer* > m_poolContainerBuffers;
      std::vector< ObjectRelationalDataBuffer* > m_poolPointerBuffers;
      std::vector< ObjectRelationalDataBuffer* > m_poolPolymorphicPointerBuffers;
      std::vector< ObjectRelationalDataBuffer* > m_dependencies;
      std::string m_updateClause;
      std::string m_idClause;
    };

  }

}

#endif
