#include "ObjectRelationalReadDataBuffer.h"
#include "RelationalStorageServiceException.h"
#include "ContainerHandlerFactory.h"
#include "IContainerHandler.h"
#include "DataElement.h"

#include "POOLCore/Context.h"
#include "POOLCore/IBlobStreamingService.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeListException.h"
#include "CoralBase/AttributeListSpecification.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "Reflex/Member.h"
#include "Reflex/Base.h"

pool::RelationalStorageService::ObjectRelationalReadDataBuffer::ObjectRelationalReadDataBuffer( coral::ISchema& schema, pool::Context* context):
  m_schema( schema ),
  m_context(context),
  m_spec( 0 ),
  m_mainDataElement(0),
  m_dataElements(),
  m_parentDataElement(0),
  m_metadata(),
  m_parentMetadata(0),
  m_parentMetadataOffset(0),
  m_table( 0 ),
  m_dictionary( 0 ),
  m_keyItems( 0 ),
  m_mapping( 0 ),
  m_keyData( false ),
  m_containerHandler( 0 ),
  m_arrayBuffers(),
  m_whereData( 0 ),
  m_whereClause(""),
  m_blobReaders(),
  m_poolContainerBuffers(),
  m_poolPointerBuffers(),
  m_poolPolymorphicPointerBuffers(),
  m_dependentObjectBuffers()
{
  m_mainDataElement = new DataElement;
}



pool::RelationalStorageService::ObjectRelationalReadDataBuffer::~ObjectRelationalReadDataBuffer()
{
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_arrayBuffers.begin();
        iBuffer != m_arrayBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_arrayBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_poolContainerBuffers.begin();
        iBuffer != m_poolContainerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolContainerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_poolPointerBuffers.begin();
        iBuffer != m_poolPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_poolPolymorphicPointerBuffers.begin();
        iBuffer != m_poolPolymorphicPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPolymorphicPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_dependentObjectBuffers.begin();
        iBuffer != m_dependentObjectBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_dependentObjectBuffers.clear();

  if(m_mainDataElement) delete m_mainDataElement;
  m_dataElements.clear();

  if ( m_whereData ) delete m_whereData;

  if ( m_containerHandler ) delete m_containerHandler;

  if ( m_spec ) m_spec->release();
  for( std::map< std::string,pool::IBlobReader* >::iterator iBlobReader=m_blobReaders.begin();
       iBlobReader!=m_blobReaders.end();
       iBlobReader++)
  {
    delete iBlobReader->second;
  }
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::build( const Reflex::Type& dict,
                                                                       const pool::ObjectRelationalMappingElement& mapping )
{
  m_spec = new coral::AttributeListSpecification;
  m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  m_whereData = new coral::AttributeList;
  // resolve possible typedef chains
  Reflex::Type dictionary = ObjectRelationalClassUtils::resolvedType( dict );
  m_dictionary = dictionary;
  m_mapping = &mapping;

  this->buildBuffer( dictionary, mapping, *m_mainDataElement, false );
  return true;
}

pool::RelationalStorageService::ObjectRelationalReadDataBuffer&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::addDependency()
{
  ObjectRelationalReadDataBuffer* newBuffer = new ObjectRelationalReadDataBuffer(m_schema, m_context);
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = m_mainDataElement;
  m_dependentObjectBuffers.push_back( newBuffer );
  return *newBuffer;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::processBaseDataMembers( const Reflex::Type& dictionary,
                                                                                        const pool::ObjectRelationalMappingElement& mapping,
                                                                                        DataElement& dataElement,
                                                                                        bool keyData ){
  int retVal = 0;
  for ( unsigned int i=0;i<dictionary.BaseSize();i++){
    Reflex::Base base = dictionary.BaseAt(i);
    Reflex::Type baseType = ObjectRelationalClassUtils::resolvedType( base.ToType() );
    retVal += processBaseDataMembers( baseType, mapping, dataElement, keyData );
    for ( unsigned int j=0;j<baseType.DataMemberSize();j++){
      Reflex::Member dataMember = baseType.DataMemberAt(j);      
      // Ignore the transients and the statics
      if ( dataMember.IsTransient() || dataMember.IsStatic() ) continue;
      // Get the member type and resolve possible typedef chains
      Reflex::Type dataMemberType = ObjectRelationalClassUtils::resolvedType( dataMember.TypeOf() );
      if ( ! dataMemberType ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for data member \"" +
                                                                                 dataMember.Name() + "\" of class \"" +
                                                                                 baseType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectRelationalReadDataBuffer::processBaseDataMembers" );
      }
      
      // check if the member is from a class in the inheritance tree
      Reflex::Type declaringType = ObjectRelationalClassUtils::resolvedType( dataMember.DeclaringType());
      std::string scope = declaringType.Name(Reflex::SCOPED|Reflex::FINAL);
      // Get the data member name
      std::string dataMemberName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( dataMember.Name(), scope );

      // Retrieve the relevant mapping element
      pool::ObjectRelationalMappingElement::const_iterator iDataMemberMapping = mapping.find( dataMemberName );
      if ( iDataMemberMapping == mapping.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Data member \"" + dataMemberName + "\" not found in the mapping element",
                                                                                 "ObjectRelationalReadDataBuffer::processBaseDataMembers" );
      }
      const pool::ObjectRelationalMappingElement& dataMemberMapping = iDataMemberMapping->second;
      DataElement& dataMemberElement = dataElement.addChild( dataMember.Offset(), base.OffsetFP() );
      retVal += this->buildBuffer( dataMemberType, dataMemberMapping, dataMemberElement, keyData );
    }
  }
  return retVal;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildBuffer( const Reflex::Type& dictionary,
                                                                             const pool::ObjectRelationalMappingElement& mapping,
                                                                             pool::RelationalStorageService::DataElement& dataElement,
                                                                             bool keyData )
{
  int retVal = 0;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Primitive ) { // Primitives
    if ( !( dictionary.IsFundamental() || dictionary.TypeInfo() == typeid(std::string) || dictionary.IsEnum() ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               "\" is not declared as primitive in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    if(m_spec->index( mapping.columnNames()[0] )==-1){
      const std::type_info* attrType = &dictionary.TypeInfo();
      if(dictionary.IsEnum()) attrType = &typeid(int);
      m_spec->extend( mapping.columnNames()[0], *attrType );
    }
    m_dataElements.push_back( &dataElement );
    ++retVal;
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Blob ) { // Blobs
    if(m_spec->index( mapping.columnNames()[0] )==-1){
      m_spec->extend<coral::Blob>( mapping.columnNames()[0] );
    }
    m_dataElements.push_back( &dataElement );
    ++retVal;
    /// book also the corresponding Blob reader
    pool::IBlobReader* blobReader = 0;
    std::map< std::string,pool::IBlobReader* >::const_iterator iBlobReader=m_blobReaders.find(mapping.columnNames()[0]);
    if( iBlobReader!=m_blobReaders.end() )
    {
      blobReader = iBlobReader->second;
    }
    if(!blobReader)
    {
      pool::IBlobStreamingService* blobStreamingService = m_context->blobStreamer();
      if ( !blobStreamingService ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot read blob variable \""+mapping.variableName()+"\": No blob streaming Service installed.","ObjectRelationalReadDataBuffer::buildBuffer" );
      }
      // fix me: version is currently not used...
      blobReader = blobStreamingService->newReader(Reflex::Type::ByName(mapping.variableType()),"");
      m_blobReaders.insert(std::make_pair(mapping.columnNames()[0],blobReader));
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Object ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::DependentObject ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) { // Objects
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
      columnNames.pop_back();
      columnNames.pop_back();
    }
    unsigned int iCol=0;
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn, ++iCol ) {
      if(m_spec->index( *iColumn )==-1){
        const std::type_info* colType = &typeid(int);
        m_spec->extend( *iColumn, *colType );
        m_whereData->extend( *iColumn, *colType  );
      }
    }

    if ( !ObjectRelationalClassUtils::isTypeObject( dictionary ) ) {
      // processing of the non-objects (top level: primitives, containers, arrays, ptr's, etc)
      ObjectRelationalMappingElement::const_iterator iArrElement = mapping.find( dictionary.Name(Reflex::SCOPED) );
      if(iArrElement==mapping.end()) throw pool::RelationalStorageService::RelationalStorageServiceException( "Mapping for array \"" + dictionary.Name(Reflex::SCOPED) +
                                                                                                              "\" has not been found.",
                                                                                                              "ObjectRelationalReadDataBuffer::buildBuffer" );
      retVal += this->buildBuffer( dictionary, iArrElement->second, dataElement, keyData );
    } else {
      if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {

        dataElement.setRefFlag();
        // Book the OID columns
        size_t index = columnNames.size();
        m_spec->extend< int >( mapping.columnNames()[index] );
        m_dataElements.push_back( &dataElement );
        ++index;
        m_spec->extend< int >( mapping.columnNames()[index] );
        m_dataElements.push_back( &dataElement );
      
      } else {
      
        // processing base classes data members
        retVal += processBaseDataMembers( dictionary, mapping, dataElement, keyData );
        // Loop over the fields of the class.
        for(unsigned int i=0;i<dictionary.DataMemberSize();i++){
          Reflex::Member dataMember = dictionary.DataMemberAt(i);
      
          // Ignore the transients and the statics
          if ( dataMember.IsTransient() || dataMember.IsStatic() ) continue;

          // Get the member type and resolve possible typedef chains
          Reflex::Type dataMemberType = ObjectRelationalClassUtils::resolvedType( dataMember.TypeOf() );
          if ( ! dataMemberType ) {
            throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for data member \"" +
                                                                                     dataMember.Name() + "\" of class \"" +
                                                                                     dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                     "ObjectRelationalReadDataBuffer::buildBuffer" );
          }

          // check if the member is from a class in the inheritance tree
          Reflex::Type declaringType = ObjectRelationalClassUtils::resolvedType( dataMember.DeclaringType());
          if( declaringType != dictionary ){
            continue;
          }
          
          // Get the field name
          std::string scope("");
          std::string dataMemberName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( dataMember.Name(), scope );

          // Retrieve the relevant mapping element
          pool::ObjectRelationalMappingElement::const_iterator iDataMemberMapping = mapping.find( dataMemberName );
          if ( iDataMemberMapping == mapping.end() ) {
            throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + dataMemberName + "\" not found in the mapping element",
                                                                                     "ObjectRelationalReadDataBuffer::buildBuffer" );
          }
          const pool::ObjectRelationalMappingElement& dataMemberMapping = iDataMemberMapping->second;
          DataElement& dataMemberElement = dataElement.addChild( dataMember.Offset(), 0 );
  
          retVal += this->buildBuffer( dataMemberType, dataMemberMapping, dataMemberElement, keyData );
        }
        
      }
    }
    
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::InlineCArray ) { // Inline c-arrays
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn ) {
      if(m_spec->index( *iColumn )==-1){
        m_spec->extend< int >( *iColumn );
      }
    }
    
      // Get the member type and resolve possible typedef chains
    Reflex::Type arrayType = ObjectRelationalClassUtils::resolvedType( dictionary.ToType() );
    if ( ! arrayType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException(
        "Missing dictionary information for the element of array \"" +dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"","ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    // Loop over the elements of the array.
    for ( unsigned int i=0;i<dictionary.ArrayLength();i++){

      // Form the element name
      std::string arrayElementLabel =  ObjectRelationalDefaultMappingRules::variableNameForArrayIndex( mapping.variableName(),i);

      // Retrieve the relevant mapping element
      pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( arrayElementLabel );
      if ( iMappingElement == mapping.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for array element \"" + arrayElementLabel + "\" not found in the mapping element",
                                                                                 "ObjectRelationalReadDataBuffer::buildBuffer" );
      }
      const pool::ObjectRelationalMappingElement& mappingElement = iMappingElement->second;
      DataElement& arrayElement = dataElement.addChild( i*arrayType.SizeOf(), 0 );

      retVal += this->buildBuffer( arrayType, mappingElement, arrayElement, keyData );
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Array ) { // Containers
    if ( !ObjectRelationalClassUtils::isTypeContainer(dictionary) ){ 
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as array in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildArrayBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolArray ) { // Pool Arrays
    if ( !ObjectRelationalClassUtils::isTypePVector(dictionary) && ! ObjectRelationalClassUtils::isTypePContainer( dictionary ) ){ 
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as array in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    if( ObjectRelationalClassUtils::isTypePVector( dictionary ) )
      this->buildArrayBuffer( dictionary, mapping, dataElement, keyData );
    else if ( ObjectRelationalClassUtils::isTypePContainer( dictionary ))
      this->buildPoolContainerBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::CArray ) { // C-arrays
    if ( ! dictionary.IsArray() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a C-array in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildCArrayBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPointer ) { // Pool pointer
    if ( ! ObjectRelationalClassUtils::isTypePoolPointer( dictionary ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as Pool Pointer in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildPoolPointerBuffer( dictionary, mapping, dataElement, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPolymorphicPointer ) { // Pool Poly Pointers
    if ( ! ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a Pool PolymorphicPointer in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    int metadataOffset = m_metadata.size();
    if(m_spec->index( mapping.columnNames()[0] )==-1){
      m_spec->extend( mapping.columnNames()[0], typeid(std::string) );
    }
    if(m_spec->index( mapping.columnNames()[1] )==-1){
      m_spec->extend( mapping.columnNames()[1], typeid(int) );
    }
    dataElement.setMetadata( metadataOffset );
    m_dataElements.push_back( &dataElement );    
    m_metadata.push_back( std::make_pair(std::string(""),0) ); // reserving a metadata slot
    ++retVal;
    this->buildPoolPolymorphicPointerBuffer( dictionary, *m_mapping, dataElement, metadataOffset, keyData );
  }
  else { // Do not treat other types yet...
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot treat anything else than objects, primitives, arrays and containers for the moment...",
                                                                             "ObjectRelationalReadDataBuffer::buildBuffer" );
  }

  return retVal;
}




bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildArrayBuffer( const Reflex::Type& dictionary,
                                                                                  const pool::ObjectRelationalMappingElement& mapping,
                                                                                  pool::RelationalStorageService::DataElement& dataElement,
                                                                                  bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_schema,
                                                                                                                                                 m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );

  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  bool posCol = true;
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    if( !posCol ){
      newBuffer->m_whereData->extend< int >( *iColumn );
      if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
      newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    }
    posCol = false;
  }

  Reflex::Type valueType;
  if( ObjectRelationalClassUtils::isTypeAssociativeContainer(dictionary)) { // associative containers.    

    Reflex::Type keyType = ObjectRelationalClassUtils::containerKeyType(dictionary);
    Reflex::Type keyResolvedType = ObjectRelationalClassUtils::resolvedType(keyType);

    if ( ! keyType ||!keyResolvedType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the key type of the container \"" +
                                                                               dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                               "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
    }

    std::string keyName = keyType.Name();
    // Retrieve the relevant mapping element
    pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( keyName );
    if ( iMappingElement == mapping.end() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + keyName + "\" not found in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
    }
    newBuffer->m_keyItems = newBuffer->buildBuffer( keyResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, true );
    valueType = ObjectRelationalClassUtils::containerDataType(dictionary);
  } else {
    valueType = ObjectRelationalClassUtils::containerValueType(dictionary);
  }

  Reflex::Type valueResolvedType = ObjectRelationalClassUtils::resolvedType(valueType);
  // Check the component type
  if ( ! valueType ||!valueResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the content type of the container \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
  }
  
  std::string valueName = valueType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( valueName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + valueName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
  }
  newBuffer->buildBuffer( valueResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildCArrayBuffer( const Reflex::Type& dictionary,
                                                                                   const pool::ObjectRelationalMappingElement& mapping,
                                                                                   pool::RelationalStorageService::DataElement& dataElement,
                                                                                   bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_schema,
                                                                                                                                                 m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );

  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  bool posCol = true;
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    if( !posCol ){
      newBuffer->m_whereData->extend< int >( *iColumn );
      if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
      newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    }
    posCol = false;
  }

  // Check the array type
  Reflex::Type arrayType = dictionary.ToType();
  Reflex::Type arrayResolvedType = ObjectRelationalClassUtils::resolvedType(arrayType);
  // Check the component type
  if ( ! arrayType || !arrayResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the element type of the array \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildCArrayBuffer" );
  }

  std::string arrayTypeName = arrayType.Name();

  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( arrayTypeName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + arrayTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildCArrayBuffer" );
  }
  newBuffer->buildBuffer( arrayResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildPoolContainerBuffer( const Reflex::Type& dictionary,
                                                                                          const pool::ObjectRelationalMappingElement& mapping,
                                                                                          DataElement& dataElement,
                                                                                          bool keyData ){
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_schema,
                                                                                                                                                 m_context);
  m_poolContainerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  Reflex::Type storeBaseType = ObjectRelationalClassUtils::containerSubType(dictionary,"store_base_type");
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( storeBaseType );
  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  bool posCol = true;
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    if( !posCol ){
      newBuffer->m_whereData->extend< int >( *iColumn );
      if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
      newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    }
    posCol = false;
  }

  Reflex::Type valueType = ObjectRelationalClassUtils::containerValueType(dictionary);
  Reflex::Type valueResolvedType = ObjectRelationalClassUtils::resolvedType(valueType);
  // Check the component type
  if ( ! valueType ||!valueResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the content type of the container \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolContainerBuffer" );
  }
  
  std::string valueName = valueType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( valueName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + valueName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolContainerBuffer" );
  }
  newBuffer->buildBuffer( valueResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );  
  return true;   
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildPoolPointerBuffer( const Reflex::Type& dictionary,
                                                                                        const pool::ObjectRelationalMappingElement& mapping,
                                                                                        pool::RelationalStorageService::DataElement& dataElement,
                                                                                        bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_schema,
                                                                                                                                                 m_context);
  m_poolPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_table = &(m_schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
  }

  // Check the  type
  Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolPointerBuffer" );
  }

  std::string ptrTypeName = ptrType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iMappingElement = mapping.find( ptrTypeName );
  if ( iMappingElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for \"" + ptrTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolPointerBuffer" );
  }
  newBuffer->buildBuffer( ptrResolvedType, iMappingElement->second, *newBuffer->m_mainDataElement, false );
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildPoolPolymorphicPointerBuffer( const Reflex::Type& dictionary,
                                                                                                   const pool::ObjectRelationalMappingElement& topLevelMapping,
                                                                                                   pool::RelationalStorageService::DataElement& dataElement,
                                                                                                   int metadataOffset,
                                                                                                   bool keyData)
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_schema,
                                                                                                                                                 m_context);
  m_poolPolymorphicPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &topLevelMapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_parentDataElement = &dataElement;
  newBuffer->m_parentMetadata = &m_metadata;
  newBuffer->m_parentMetadataOffset = metadataOffset;
  newBuffer->m_keyData = keyData;

  // Check the  type
  ROOT::Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  ROOT::Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolPointerBuffer" );
  }
  return true;
}

const coral::AttributeListSpecification&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::dataSpecification() const
{
  return *m_spec;
}

const std::vector<pool::RelationalStorageService::DataElement*>&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::dataElements() const
{
  return m_dataElements;
}

pool::RelationalStorageService::DataElement&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::parentDataElement() const
{
  if(!m_parentDataElement) throw RelationalStorageServiceException("Top level buffer does not have parent DataElement.",
                                                                   "ObjectRelationalReadDataBuffer::parentDataElement");
  return *m_parentDataElement;
}

const std::vector< std::pair<std::string,int> >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::metadata() const {
  return m_metadata;
}

std::vector< std::pair<std::string,int> >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::metadata() {
  return m_metadata;
}

const std::vector< std::pair<std::string,int> >*
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::parentMetadata() const{
  return m_parentMetadata;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::parentMetadataOffset() const{
  return m_parentMetadataOffset;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::keyItems() const
{
  return m_keyItems;
}

coral::ITable&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::table()
{
  return *m_table;
}

Reflex::Type
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::dictionary() const
{
  return m_dictionary;
}


const pool::ObjectRelationalMappingElement&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::mapping() const
{
  return *m_mapping;
}


pool::RelationalStorageService::IContainerHandler&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::containerHandler()
{
  return *m_containerHandler;
}


bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::isKeyData() const
{
  return m_keyData;
}


pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::begin()
{
  return m_arrayBuffers.begin();
}

pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::end()
{
  return m_arrayBuffers.end();
}

const std::vector<pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::poolContainerBuffers() const{
  return m_poolContainerBuffers;
}

const std::vector<pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::poolPointerBuffers() const{
  return m_poolPointerBuffers;
}

const std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::poolPolymorphicPointerBuffers() const {
  return m_poolPolymorphicPointerBuffers;
}

const std::string&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::whereClauseForId() const
{
  return m_whereClause;
}


coral::AttributeList&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::whereDataForId()
{
  return *m_whereData;
}

void
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::readBlob(const std::string& variableName, const coral::Blob& data, void* startAddress)
{
  pool::IBlobReader* blobReader = 0;
  std::map< std::string,pool::IBlobReader* >::const_iterator iBlobReader=m_blobReaders.find(variableName);
  if( iBlobReader!=m_blobReaders.end() )
  {
    blobReader = iBlobReader->second;
  }
  if(!blobReader)
  {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot read from blob: The reader has not been booked.","ObjectRelationalReadDataBuffer::writeBlob" );

  }
  blobReader->read(data,startAddress);
}
