#ifndef OBJECTRELATIONALREADDATABUFFER_H
#define OBJECTRELATIONALREADDATABUFFER_H

#include <vector>
#include <map>
#include <string>
#include "Reflex/Type.h"
#include "RVersion.h"

namespace coral {

  class ITable;
  class ISchema;
  class AttributeList;
  class AttributeListSpecification;
  class Blob;
}

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif
 

namespace pool {
  class ObjectRelationalMappingElement;
  class IBlobReader;
  class Context;

  namespace RelationalStorageService {

    class IContainerHandler;
    class DataElement;

    class ObjectRelationalReadDataBuffer {
    public:
      /// Constuctor
      explicit ObjectRelationalReadDataBuffer(coral::ISchema& schema, pool::Context* context);

      /// Destructor
      ~ObjectRelationalReadDataBuffer();

      /// Builds the structure given a mapping element and the dictionary information
      bool build( const Reflex::Type& dictionary,
                  const pool::ObjectRelationalMappingElement& mapping );

      /// adds a new buffer to the dependency list
      ObjectRelationalReadDataBuffer& addDependency();

      /// Returns the data specification
      const coral::AttributeListSpecification& dataSpecification() const;

      /// Returns the children data elements
      const std::vector<DataElement*>& dataElements() const;

      /// Returns the parent data elements
      DataElement& parentDataElement() const;

      /// Returns the metadata vector
      const std::vector< std::pair<std::string,int> >& metadata() const;
      
      /// Returns the metadata vector
      std::vector< std::pair<std::string,int> >& metadata();

      /// Returns the parent metadata
      const std::vector< std::pair<std::string,int> >* parentMetadata() const;
      
      /// Returns the parent metadata offset
      int parentMetadataOffset() const;
      
      /// Returns the number of items belonging to the key (in case of an associative container)
      int keyItems() const;

      /// Returns the associoted table handle
      coral::ITable& table();

      /// Returns the associated dictionary information
      Reflex::Type dictionary() const;

      /// Returns the relevant mapping element
      const pool::ObjectRelationalMappingElement& mapping() const;

      /// Returns the container handler
      IContainerHandler& containerHandler();

      /// Returns a flag indicating whether the starting address is the key of an associative container
      bool isKeyData() const;

      /// Returns the where clause for pinning the id
      const std::string& whereClauseForId() const;

      /// Returns the data for the where clause.
      coral::AttributeList& whereDataForId();

      /// Returns the original version of the specified data, using the blob reader registered in the pool context 
      void readBlob(const std::string& variableName, const coral::Blob& data, void* startAddress);

      public:
      typedef std::vector< ObjectRelationalReadDataBuffer* >::iterator iterator;
      iterator begin();
      iterator end();
      const std::vector<ObjectRelationalReadDataBuffer* >& poolContainerBuffers() const;
      const std::vector<ObjectRelationalReadDataBuffer* >& poolPointerBuffers() const;
      const std::vector< ObjectRelationalReadDataBuffer* >& poolPolymorphicPointerBuffers() const;

    private:
      /// process the data members from the base trees 
      int processBaseDataMembers( const Reflex::Type& dictionary,
                                  const pool::ObjectRelationalMappingElement& mapping,
                                  DataElement& dataElement,
                                  bool keyData );
      /// builds a buffer entry
      int buildBuffer( const Reflex::Type& dictionary,
                       const pool::ObjectRelationalMappingElement& mapping,
                       DataElement& dataElement,
                       bool keyData );

      /// builds an array buffer
      bool buildArrayBuffer( const Reflex::Type& dictionary,
                             const pool::ObjectRelationalMappingElement& mapping,
                             DataElement& dataElement,
                             bool keyData );

      /// builds an array buffer
      bool buildCArrayBuffer( const Reflex::Type& dictionary,
                              const pool::ObjectRelationalMappingElement& mapping,
                              DataElement& dataElement,
                              bool keyData );

      /// builds an array buffer
      bool buildPoolContainerBuffer( const Reflex::Type& dictionary,
                                     const pool::ObjectRelationalMappingElement& mapping,
                                     DataElement& dataElement,
                                     bool keyData );

      /// builds a pool ptr buffer
      bool buildPoolPointerBuffer( const Reflex::Type& dictionary,
                                   const pool::ObjectRelationalMappingElement& mapping,
                                   DataElement& dataElement,
                                   bool keyData );
      
      /// builds a Pool PolyPtr buffer
      bool buildPoolPolymorphicPointerBuffer( const Reflex::Type& dictionary,
                                              const pool::ObjectRelationalMappingElement& topLevelMapping,
                                              DataElement& dataElement,
                                              int metadataOffset,
                                              bool keyData);
      private:
      
      coral::ISchema& m_schema;
      pool::Context* m_context;
      coral::AttributeListSpecification* m_spec;
      DataElement* m_mainDataElement;
      std::vector<DataElement*> m_dataElements;
      DataElement* m_parentDataElement;
      std::vector< std::pair<std::string,int> > m_metadata;
      std::vector< std::pair<std::string,int> >*  m_parentMetadata;
      int m_parentMetadataOffset;
      coral::ITable* m_table;
      Reflex::Type m_dictionary;
      int m_keyItems;
      const pool::ObjectRelationalMappingElement* m_mapping;
      bool m_keyData;
      IContainerHandler* m_containerHandler;
      std::vector< ObjectRelationalReadDataBuffer* > m_arrayBuffers;
      coral::AttributeList* m_whereData;
      std::string m_whereClause;
      std::map< std::string, pool::IBlobReader* > m_blobReaders;
      std::vector< ObjectRelationalReadDataBuffer* > m_poolContainerBuffers;
      std::vector< ObjectRelationalReadDataBuffer* > m_poolPointerBuffers;
      std::vector< ObjectRelationalReadDataBuffer* > m_poolPolymorphicPointerBuffers;
      std::vector< ObjectRelationalReadDataBuffer* > m_dependentObjectBuffers;
    };

  }

}

#endif
