#ifndef ROOTSTORAGE_DBDATABASEMERGER_H
#define ROOTSTORAGE_DBDATABASEMERGER_H

// C++ includes 
#include <string>
#include <vector>
#include <map>

// Forward declarations
class TTree;
class TFile;
class TBranch;


/*
 * POOL namespace declaration
 */
namespace pool {

  typedef int DbStatus;
  enum { ERROR=0, SUCCESS=1 };

  struct DbContainerSection {
    int start;
    int length;
    int offset;
  };

  /**@class DbDatabaseMerger DbDatabaseMerger.h StorageSvc/DbDatabaseMerger.h
   *
   *  @author  M.Frank
   *  @version 1.0
   */
  class DbDatabaseMerger {

    typedef std::vector<DbContainerSection>           DbContainerSections;
    typedef std::map<std::string,DbContainerSections> DbDatabaseSections;

    DbDatabaseSections m_sections;
    TFile*             m_output;
    TTree*             m_sectionTree;
    TBranch*           m_sectionBranch;

  public:
    /// Standard constructor
    DbDatabaseMerger();
    /// Default destructor
    virtual ~DbDatabaseMerger();
    /// Check if a database exists
    bool exists(const std::string& fid) const;
    /// Create new output file
    DbStatus create(const std::string& fid);
    /// Attach to existing output file for further merging
    DbStatus attach(const std::string& fid);
    /// Close output file
    DbStatus close();
    /// Merge new input to existing output
    DbStatus merge(const std::string& fid);
    /// Dump collected database sections
    void dumpSections();
    /// Save new sections to the output file
    DbStatus saveSections();
    DbStatus createFID();
  };
}    // End namespace pool

#endif // ROOTSTORAGE_DBDATABASEMERGER_H


#ifndef ROOTSTORAGE_DBDATABASEMERGER_H
#include "DbDatabaseMerger.h"
#endif // ROOTSTORAGE_DBDATABASEMERGER_H
#include "TUUID.h"
#include "TFile.h"
#include "TTree.h"
#include "TBranch.h"
#include "TKey.h"
#include "TROOT.h"
#include "TSystem.h"
#include "TTreeCloner.h"
#include <iostream>
#include <iomanip>


using namespace pool;
using namespace std;

namespace {
  const char* getLinkContainer(char* txt) {
    char* p = ::strstr(txt,"[CNT=");
    if ( p ) {
      char* q = strchr(p+4,']');
      *q = 0;
      return p+5;
    }
    return 0;
  }
  string getRootContainer(const char* t) {
    string r;
    for(const char* p=t; *p; ++p) {
      r += (*p == '/') ? '_' : *p;
    }
    return r;
  }
}

/// Standard constructor
DbDatabaseMerger::DbDatabaseMerger() : m_output(0) {
}

/// Default destructor
DbDatabaseMerger::~DbDatabaseMerger() {
  close();
}

/// Check if a database exists
bool DbDatabaseMerger::exists(const std::string& fid) const {
  Bool_t result = gSystem->AccessPathName(fid.c_str(), kFileExists);
  if ( result == kFALSE )  
    cout << "file " << fid << " EXISTS!" << endl;
  else
    cout << "file " << fid << " DOES NOT EXIST!" << endl;
  return result == kFALSE;
}

/// Attach to existing output file for further merging
DbStatus DbDatabaseMerger::attach(const string& fid) {
  if ( m_output ) {
    cout << "+++ Another output file " << m_output->GetName() << " is already open. Request denied." << endl;
    return ERROR;
  }
  else if ( !exists(fid) ) {
    cout << "+++ Cannot attach output file " << fid << " --- file does not exist." << endl;
    return ERROR;
  }
  m_output = TFile::Open(fid.c_str(),"UPDATE");
  if ( m_output && !m_output->IsZombie() ) {
    m_sectionTree = (TTree*)m_output->Get("##Sections");
    if ( 0 == m_sectionTree ) {
      close();
      cout << "+++ Existing output file " << fid << " is no merge file --- Request denied." << endl;
      delete m_output;
      m_output = 0;
      return ERROR;
    }
    m_sectionBranch = m_sectionTree->GetBranch("db_string");
    cout << "+++ Opened new output file " << fid << "." << endl;
    return SUCCESS;
  }
  cout << "+++ Failed to open new output file " << fid << "." << endl;
  return ERROR;
}

/// Create new output file
DbStatus DbDatabaseMerger::create(const string& fid) {
  if ( m_output ) {
    cout << "+++ Another output file " << m_output->GetName() << " is already open. Request denied." << endl;
    return ERROR;
  }
  else if ( exists(fid) ) {
    cout << "+++ Cannot create output file " << fid << " --- file already exists." << endl;
    return ERROR;
  }
  m_output = TFile::Open(fid.c_str(),"RECREATE");
  if ( m_output && !m_output->IsZombie() )     {
    m_sectionTree   = new TTree("##Sections","Tree with database section data");
    m_sectionBranch = m_sectionTree->Branch("db_string",0,"db_string/C",32*1024);
    cout << "+++ Opened new output file " << fid << "." << endl;
    return SUCCESS;
  }
  cout << "+++ Failed to open new output file " << fid << "." << endl;
  return ERROR;
}

/// Close output file
DbStatus DbDatabaseMerger::createFID() {
  if ( m_output ) {
    TTree* t = (TTree*)m_output->Get("##Params");
    if ( t ) {
      char text[132];
      unsigned char buf[16];
      TUUID uuid;
      TBranch* b = t->GetBranch("db_string");
      uuid.GetUUID(buf);
      sprintf(text,"[NAME=FID][VALUE=%08X-%04hX-%04hX-%02hhX%02hhX-%02hhX%02hhX%02hhX%02hhX%02hhX%02hhX]",
	     (*(int*)buf),*(short*)(buf+4),*(short*)(buf+6),
	      *(buf+8),*(buf+8),
	      *(buf+10),*(buf+11),
	      *(buf+12),*(buf+13),
	      *(buf+14),*(buf+15));
      b->SetAddress(text);
      t->Fill();
      t->Write();
      cout << "+++ Added new GUID " << text << " to merge file." << endl;
      return SUCCESS;
    }
  }
  cout << "+++ Failed to add new GUID to merge file." << endl;
  return ERROR;
}

/// Close output file
DbStatus DbDatabaseMerger::close() {
  if ( m_output ) {
    if ( m_sectionTree ) m_sectionTree->Write();
    m_output->Write();
    m_output->Purge();
    m_output->ls();
    m_output->Close();
    delete m_output;
    m_output = 0;
    m_sectionTree = 0;
    m_sectionBranch = 0;
  }
  return SUCCESS;
}

/// Save new sections to the output file
DbStatus DbDatabaseMerger::saveSections()     {
  if ( m_sectionBranch ) {
    int total = 0;
    int nbytes = 0;
    char text[1024];
    m_sectionBranch->SetAddress(text);
    for(DbDatabaseSections::const_iterator i=m_sections.begin(); i!= m_sections.end();++i) {
      int cnt = 0;
      string cont = (*i).first;
      const DbContainerSections& cntSects = (*i).second;
      for(DbContainerSections::const_iterator j=cntSects.begin(); j != cntSects.end();++j, ++cnt) {
	::sprintf(text,"[CNT=%s][OFF=%d][START=%d][LEN=%d]",
		  cont.c_str(),(*j).offset,(*j).start,(*j).length); 
	int nb = m_sectionTree->Fill();
	++total;
	if ( nb > 0 ) {
	  nbytes += nb;
	}
	else {
	  cout << "+++ Failed to update ##Sections tree with new entries." << endl;
	}
      }
    }
    cout << "+++ Added " << total << " ##Sections entries with " << nbytes << " bytes in total." << endl;
    TTree *links = (TTree*)m_output->Get("##Links");
    Long64_t len = links ? links->GetEntries() : 0;
    TBranch* blinks = links->GetBranch("db_string");
    blinks->SetAddress(text);
    for(Long64_t i=0; i<len; ++i) {
      links->GetEntry(i);
      if ( strstr(text,"CNT=##Sections") != 0 ) {
	return SUCCESS;
      }
    }
    ::sprintf(text,"[DB=<localDB>][CNT=##Sections][CLID=DA8F479C-09BC-49D4-94BC-99D025A23A3B][TECH=00000200][OID=%08X-FFFFFFFF]",
	      int(len+2));
    cout << "Updated bytes on ##Links:" << links->Fill() << endl;
    links->Write();
    return SUCCESS;
  }
  return ERROR;
}

/// Dump collected database sections
void DbDatabaseMerger::dumpSections() {
  for(DbDatabaseSections::const_iterator i=m_sections.begin(); i!= m_sections.end();++i) {
    int cnt = 0;
    string prefix = (*i).first;
    const DbContainerSections& cntSects = (*i).second;
    for(DbContainerSections::const_iterator j=cntSects.begin(); j != cntSects.end();++j, ++cnt) {
      char text[32];
      ::sprintf(text,"'][%d]",cnt); 
      cout << "+++ " << setw(60) << left << "section['"+prefix+text
	   << "  Link offset:" << setw(5) << right << (*j).offset 
	   << "  Start:" << setw(8) << right << (*j).start 
	   << " ... "    << setw(8) << right << (*j).start+(*j).length 
	   << " ["       << (*j).length << "  entries]." << endl;
    }
  }
}

/// Merge new input to existing output
DbStatus DbDatabaseMerger::merge(const string& fid) {
  if ( m_output ) {
    TFile* source = TFile::Open(fid.c_str());
    if ( source && !source->IsZombie() ) {
      char text[1024];
      TTree *links = (TTree*)m_output->Get("##Links");
      Long64_t lnk_offset = links ? links->GetEntries() : 0;
      TTree* src_links = (TTree*)source->Get("##Links");
      TBranch* src_blinks = src_links->GetBranch("db_string");
      src_blinks->SetAddress(text);
      map<string,string> redirect;
      redirect["##Links"]  = "##Links";
      redirect["##Shapes"] = "##Shapes";
      for(size_t i=0, n=src_links->GetEntries(); i<n; ++i) {
	src_links->GetEntry(i);
	cout << text << endl;
	const char* l = getLinkContainer(text);
	if ( l ) redirect[getRootContainer(l)] = l;
      }
      m_output->cd();
      TIter nextkey(source->GetListOfKeys());
      for(TKey* key = (TKey*)nextkey(); key; key = (TKey*)nextkey() ) {
	const char *classname = key->GetClassName();
	TClass *cl = gROOT->GetClass(classname);
	if (!cl) continue;
	if (cl->InheritsFrom("TTree")) {
	  string name = key->GetName();
	  source->cd();
	  TTree *src_tree = (TTree*)source->Get(key->GetName());
	  Long64_t src_entries = src_tree->GetEntries();
	  m_output->cd();
	  DbContainerSection s;
	  s.start = 0;
	  s.length = src_entries;
	  s.offset = lnk_offset;

	  TTree *out_tree = (TTree*)m_output->Get(key->GetName());
	  if ( out_tree == 0 ) {
	    out_tree = src_tree->CloneTree(-1,"fast");
	    cout << "+++ Created new Tree " << out_tree->GetName() << endl;
	    out_tree->Write();
	  }
	  else {
	    m_output->GetObject(key->GetName(),out_tree);
	    TTreeCloner cloner(src_tree,out_tree,"fast");//,TTreeCloner::kNoWarnings);
	    if (cloner.IsValid()) {
	      Long64_t out_entries = out_tree->GetEntries();
	      s.start = out_entries;
	      out_tree->SetEntries(out_entries+src_entries);
	      Bool_t res = cloner.Exec();
	      cout << "+++ Merged tree: " << out_tree->GetName() << " res=" << res << endl;
	      out_tree->Write();
	    }
	    else {
	      // Fast cloning is not possible for this input TTree.
	      // ... see TTree::CloneTree for example of recovery code ...
	      cout << "+++ Got a tree where fast cloning is not possible -- operation failed." << endl;
	      continue;
	    }
	  }
	  map<string,string>::const_iterator ir=redirect.find(name);
	  if ( ir != redirect.end() ) {
	    m_sections[(*ir).second].push_back(s);
	  }
	}
	else   {
	  cout << "+++ Ignore key " << key->GetName() << endl;
	}
      }
      delete source;
      m_output->SaveSelf(kTRUE);
      m_output->Purge();
      m_output->cd();
      return SUCCESS;
    }
    cout << "+++ Cannot open input file:" << source << endl;
    m_output->cd();
    return ERROR;
  }
  cout << "+++ No valid output file present. Merge request refused for fid:" << fid << endl;
  return ERROR;
}

static bool s_dbg = false;

int merge(const char* target, const char* source, bool fixup=false, bool dbg=true) {
  s_dbg = dbg;
  gSystem->Load("libCintex");
  DbDatabaseMerger m;
  DbStatus ret = m.exists(target) ? m.attach(target) : m.create(target);
  if ( ret == SUCCESS ) {
    ret = m.merge(source);
    if ( ret == SUCCESS ) {
      m.dumpSections();
      if ( fixup ) m.createFID();
      m.saveSections();
      return m.close();
    }
  }
  cout << "+++ Cannot open output file:" << target << endl;
  return 0;
}
