/**
 * @file CollListAttrib.cpp
 * @brief Utility to list the metadata of a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollListAttrib.cpp,v 1.27 2009/05/13 15:32:22 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"

#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"

#include "Args2Container.h"
#include "CmdLineArgs2.h"
#include "ArgQual.h"
#include "SrcInfo.h"
#include "CatalogInfo.h"

#include <iostream>
#include <iomanip>
#include <fstream>

using namespace std;


void dumpTokens(const pool::ICollectionDescription& description); 
void dumpAttributes(const pool::ICollectionDescription& description); 

void dumpOverlapEval(const pool::ICollectionDescription& desc0,
                     const pool::ICollectionDescription& desc1,
                     const vector< pair<string,string> >& spec0,
                     const vector< pair<string,string> >& spec1,
                     const string spectype);


int main(int argc, const char *argv[])
{
   string thisProgram("CollListAttrib.exe");
   vector<std::string> argv_v;
   for (int i=0; i<argc; ++ i) argv_v.push_back(string(argv[i]));

   try  { 
      pool::CollectionService   collectionService;
      coral::MessageStream log( thisProgram );

      Args2Container argsVec( thisProgram, true, &log );
      argsVec.desc << "Lists the attributes of a collection "
                   << "or pair of collections. " << endl << "If it is a pair, it will also "
                   << "tell which attributes are shared and which are not. " << endl
                   << "It will also print this information to files." ;

      // Classes with shared cli keys and their argument properties
      // Add them to the list
      // SrcInfo : input collections
      // CatInfo : collection catalogs
      CatalogInfo catinfo; argsVec.push_back(&catinfo); 
      SrcInfo     srcinfo; argsVec.push_back(&srcinfo);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;

      if (srcinfo.nSrc()>2) {
	 cerr << "This program takes a max of 2 input collections" << endl;
	 cerr << "There were " << srcinfo.nSrc() << " inputs collections on the command line" << endl;
	 return 1;
      }
      
      catinfo.setCatalogs( &collectionService );
  

      unsigned int collCounter = 0;
      std::string attribName = "";
      std::string attribTypeName = "";
	
      map<int, pool::ICollection*> collMap;
      map<int, vector<pair<string,string> > > tokSpecs;
      map< int, vector< pair<string,string> > > attSpecs;
	 
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ )  {
	 bool readOnly( true );
	 pool::ICollection* collection;
	 try {
	    collection = collectionService.handle( srcinfo.name(i), srcinfo.type(i), srcinfo.connect(), readOnly );
	 }
	 catch ( pool::Exception& poolException ) {
	    std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
	    cout << "Unable to open collection " << srcinfo.name(i) << endl;
	    cout << "---> STOPPING HERE" << endl;
	    return 5;
	 }

	 const pool::ICollectionDescription &description = collection->description();
	 collMap[i] = collection;
	 collCounter++;
	 vector< pair<string,string> > spec;
	 for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	    for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
	       const pool::ICollectionColumn&   column = description.attributeColumn(c, f);
	       spec.push_back(make_pair(column.name(),column.type()));
	    }
	 }
	 attSpecs[i]=spec;
	 spec.clear();
	 for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	    for( int c=0; c < description.numberOfTokenColumns( f ); c++ ) {
	       const pool::ICollectionColumn&   column = description.tokenColumn(c, f);
	       spec.push_back(make_pair(column.name(),column.type()));
	    }
	 }
	 tokSpecs[i]=spec;
      }


      std::cout << "--------------------------------------------------------------" << std::endl;
      std::cout << "Collection list: " << endl;
      for (unsigned int i=0; i < collCounter; ++i) {
	 std::cout << "NAME: ";
	 std::cout << collMap[i]->description().name();
	 std::cout << "   TYPE: ";
	 std::cout << collMap[i]->description().type();
	 std::cout << "   NFRAG: ";
	 std::cout << collMap[i]->description().numberOfCollectionFragments();
	 std::cout << std::endl;
      }
      
      if (collCounter==1) {
	 // For a single input, just dump the description
	 dumpTokens(collMap[0]->description());
	 dumpAttributes(collMap[0]->description());
      }
      else if (collCounter == 2 && 
	       collMap[0]->description().equals(collMap[1]->description())
	       ) {
	 // If the two have identical descriptions, then dump the first one
	 std::cout << "--------------------------------------------------------------" << std::endl;
	 std::cout << " ***" << std::endl;
	 std::cout << " *** Tokens and Attributes IDENTICAL for these Collections ***" << std::endl;
	 std::cout << " ***" << std::endl;
	 dumpTokens(collMap[0]->description());
	 dumpAttributes(collMap[0]->description());
      }
      else {
	 // If they differ, then 
	 // -- First, Look at token differences
	 //    -- If not any, then dump tokens
	 //    -- If differences exist, dump common tokens, then differences
	 //    -- also write 3 files: common.tokens, <coll1>Only.tokens, <coll2>Only.tokens
	 // -- Next, look at attribute differences
	 //    -- If not any, then dump attributes
	 //    -- If differences exist, dump common attributes, then differences
	 //    -- also write 3 files: common.atts, <coll1>Only.atts, <coll2>Only.atts

	 stable_sort(tokSpecs[0].begin(),tokSpecs[0].end());
	 stable_sort(tokSpecs[1].begin(),tokSpecs[1].end());
	 stable_sort(attSpecs[0].begin(),attSpecs[0].end());
	 stable_sort(attSpecs[1].begin(),attSpecs[1].end());
	 // Look at token intersection first
	 dumpOverlapEval(collMap[0]->description(),
			 collMap[1]->description(),
			 tokSpecs[0],
			 tokSpecs[1],
			 "Token");
	 // Look at attribute intersection second
	 dumpOverlapEval(collMap[0]->description(),
			 collMap[1]->description(),
			 attSpecs[0],
			 attSpecs[1],
			 "Attribute");
      }
      std::cout << "---------------------------------------------------------" << std::endl;
      std::cout << std::endl;
      std::cout << "Number of collections scanned is: " 
		<< collCounter << std::endl;
      std::cout << std::endl;

      for( unsigned int i=0; i < srcinfo.nSrc(); i++ ) {      
	 collMap[i]->close();
      }

      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::exception: " << exception.what() << std::endl;;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception caught in main()." << std::endl;;
      return 1;
   }
}



std::pair<unsigned int,unsigned int>
getColumnWidths(const pool::ICollectionDescription& description, bool tokens=false)
{
   unsigned int maxNameSize=0,maxTypeNameSize=0;
   if (tokens) {
      for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	 for( int c=0; c < description.numberOfTokenColumns( f ); c++ ) {
	    const pool::ICollectionColumn& column = description.tokenColumn(c, f);
	    if( maxNameSize < column.name().size() )
	       maxNameSize = column.name().size();
	    if( maxTypeNameSize < column.type().size() )
	       maxTypeNameSize = column.type().size();
	 }
      }
   }
   else {
      for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	 for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
	    const pool::ICollectionColumn& column = description.attributeColumn(c, f);
	    if( maxNameSize < column.name().size() )
	       maxNameSize = column.name().size();
	    if( maxTypeNameSize < column.type().size() )
	       maxTypeNameSize = column.type().size();
	 }
      }
   }
   return std::make_pair(maxNameSize,maxTypeNameSize);
}

void dumpTokens(const pool::ICollectionDescription& description) 
{
   string spectype("Token");
   std::pair<unsigned int,unsigned int> maxpair = getColumnWidths(description,true);
   unsigned int maxNameSize = maxpair.first;
   unsigned int maxTypeNameSize = maxpair.second;
   std::cout << "--------------------------------------------------------------" << std::endl;
   std::cout << "Number of " << spectype << "s is: " 
	     << description.numberOfTokenColumns() << std::endl;
   std::cout << spectype << "s are: " << std::endl;
   for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
      for( int c=0; c < description.numberOfTokenColumns( f ); c++ ) {
	 const pool::ICollectionColumn&	column = description.tokenColumn(c, f);
	 std::cout.setf(std::ios::left);
	 std::cout << " NAME: ";
	 std::cout.width(maxNameSize+5);
	 std::cout << column.name();
	 std::cout << " TYPE: ";
	 std::cout.width(maxTypeNameSize+5);
	 std::cout << column.type();
	 std::cout << " INFO: ";
	 std::cout << column.annotation();
	 std::cout << std::endl;
      }
   }
}

void dumpAttributes(const pool::ICollectionDescription& description)
{
   string spectype("Attribute");
   std::pair<unsigned int,unsigned int> maxpair = getColumnWidths(description);
   unsigned int maxNameSize = maxpair.first;
   unsigned int maxTypeNameSize = maxpair.second;
   std::cout << "--------------------------------------------------------------" << std::endl;
   std::cout << "Number of " << spectype << "s is: " 
	     << description.numberOfAttributeColumns() << std::endl;
   std::cout << spectype << "s are: " << std::endl;
   for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
      for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
	 const pool::ICollectionColumn&	column = description.attributeColumn(c, f);
	 std::cout.setf(std::ios::left);
	 std::cout << " NAME: ";
	 std::cout.width(maxNameSize+5);
	 std::cout << column.name();
	 std::cout << " TYPE: ";
	 std::cout.width(maxTypeNameSize+5);
	 std::cout << column.type();
	 std::cout << " INFO: ";
	 std::cout << column.annotation();
	 std::cout << std::endl;
      }
   }
}

void dumpOverlapEval(const pool::ICollectionDescription& desc0,
		     const pool::ICollectionDescription& desc1,
		     const vector< pair<string,string> >& spec0,
		     const vector< pair<string,string> >& spec1,
		     const string spectype)
{
   // First check that inputs are valid
   vector< vector< pair<string,string> > > Specs;
   Specs.push_back(spec0);
   Specs.push_back(spec1);
   vector<string> Descs;
   Descs.push_back(desc0.name());
   Descs.push_back(desc1.name());
   vector< pair<string,string> > tmp(max(spec0.size(),spec1.size()));
   vector< pair<string,string> >::iterator last = 
      set_intersection( Specs[0].begin(),Specs[0].end(),
			Specs[1].begin(),Specs[1].end(),
			tmp.begin() );
   vector< pair<string,string> > output(tmp.begin(),last); 
   unsigned int intersize = output.size();
   if (intersize==Specs[0].size() && 
       Specs[0].size()==Specs[1].size()) {
      std::cout << "--------------------------------------------------------------" << std::endl;
      std::cout << " ***" << std::endl;
      std::cout << " *** " << spectype << "s IDENTICAL for these Collections ***" << std::endl;
      std::cout << " ***" << std::endl;
      if (spectype=="Token") dumpTokens(desc0);
      else if (spectype=="Attribute") dumpAttributes(desc0); 
      else cerr << "dumpOverlapEval ERROR: unknown spectype = " << spectype << endl; 
   }
   else {
      std::cout << "--------------------------------------------------------------" << std::endl;
      std::cout << " ***" << std::endl;
      std::cout << " *** " << spectype << "s DIFFER for these Collections ***" << std::endl;
      std::cout << " ***" << std::endl;
      std::cout << "--------------------------------------------------------------" << std::endl;
      std::cout << "Number of Common " << spectype << " Names is: " 
		<< output.size() << std::endl;
      // First dump intersection
      std::cout << "Common " << spectype << " Names are: " << std::endl;
      std::ofstream intersect_names;
      std::string name("intersect_names."+spectype);
      intersect_names.open(name.c_str(),ios::out | ios::trunc);
      for( vector< pair<string,string> >::iterator c = output.begin(); 
	   c < output.end(); 
	   c++ ) {
	 std::cout << " NAME: ";
	 std::cout << c->first;
	 std::cout << "   TYPE: ";
	 std::cout << c->second;
	 std::cout << std::endl;
	 intersect_names << c->first << std::endl;
      }
      intersect_names.close();
      for (int i=0; i<2; ++i) {
	 int j;
	 if (i==0) j=1; else j=0;
	 std::ofstream names;
	 if (Specs[i].size() > intersize) {
	    std::cout << "Number of " << spectype << " Names Specific to "
		      << Descs[i] << " is: " 
		      << Specs[i].size() - intersize << std::endl;
	    std::cout << spectype << " Names are: " << std::endl;
	    std::string name(Descs[i]+"Only."+spectype);
	    names.open(name.c_str(), ios::out | ios::trunc);
	    vector< pair<string,string> > tmp(max(Specs[0].size(),Specs[1].size()));
	    vector< pair<string,string> >::iterator last2 = 
	       set_difference(Specs[i].begin(),Specs[i].end(),
			      Specs[j].begin(),Specs[j].end(),
			      tmp.begin());
	    vector< pair<string,string> > solo(tmp.begin(),last2); 
	    for( vector< pair<string,string> >::iterator c = solo.begin(); 
		 c < solo.end(); 
		 c++ ) {
	       std::cout << " NAME: ";
	       std::cout << c->first;
	       std::cout << "   TYPE: ";
	       std::cout << c->second;
	       std::cout << std::endl;
	       names << c->first << std::endl;
	    }
	 }
	 names.close();
      }
   }
}

