/**
 * @file CollListFileGUID.cpp
 * @brief Utility to list the file GUID's used by a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollListFileGUID.cpp,v 1.27 2009/09/17 15:12:25 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"

#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ITransaction.h"

#include "CoralBase/MessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "Args2Container.h"
#include "SrcInfo.h"
#include "QueryInfo.h"
#include "CatalogInfo.h"

#include <iostream>
using namespace std;


int main(int argc, const char *argv[])
{
   string thisProgram("CollListFileGUID");

   // Convert argv to vector of strings
   vector<string> argv_v;
   for( int i=0; i<argc; ++ i )
      argv_v.push_back( argv[i] );

   try
   { 
      coral::MessageStream log( thisProgram );
      pool::CollectionService   collectionService;

      // vector of CmdLineArg objects
      Args2Container argsVec(thisProgram, true, &log);
      argsVec.desc << thisProgram
                   << " is a tool for listing POOL file GUIDs referenced from the source event collection(s). "
                   << endl;

      QualList markers;
      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      argsVec.push_back(&cmdLineArgs); 

      CatalogInfo catinfo; argsVec.push_back(&catinfo);
      QueryInfo queryinfo; argsVec.push_back(&queryinfo);
      SrcInfo srcinfo;     argsVec.push_back(&srcinfo);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;

      // Fill appropriate vectors based on *specific* cmdLineArgs
      map < string, pair <int,int> >::iterator mit=cmdLineArgs.begin();
      while( mit != cmdLineArgs.end() ) {
         string qual = mit->first;
         //int ifirst = mit->second.first;
         log << coral::Error << "Unrecognized option: " << qual << endl
             << "Use -help for option list" << coral::MessageStream::endmsg;
         exit(-1); 
         ++mit;
      } 

      catinfo.setCatalogs( &collectionService );

      std::map<int, pool::ICollection*> collMap;
      std::map< std::string, int > fileIdMap;
      int fileCounter = 0;
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {

	 bool readOnly( true );
         pool::ICollection *collection = collectionService.handle( srcinfo.name(i), srcinfo.type(i), srcinfo.connect(), readOnly );
	 collMap[i] = collection;
	 
	 log << coral::Debug
             << ">> Creating query for the collection" << coral::MessageStream::endmsg;            
	 pool::ICollectionQuery *collQuery = collection->newQuery();
         collQuery->setCondition( queryinfo.query() );
         if( queryinfo.queryOptions().size() ) {
            collQuery->addToOutputList( queryinfo.queryOptions() );
         } else {
	    collQuery->selectAllTokens();
         }
	 log << coral::Debug
	     << ">> Executing the query " << coral::MessageStream::endmsg;
	 pool::ICollectionCursor& cursor = collQuery->execute();

	 while( cursor.next() ) {
	    const pool::TokenList &tokens = cursor.currentRow().tokenList();
	    for( pool::TokenList::const_iterator iter = tokens.begin();
		 iter != tokens.end(); ++iter ) {
	       std::string fileId = iter->dbID();
	       std::map< std::string, int >::iterator mapIter;
	       mapIter = fileIdMap.find( fileId );
	       if( mapIter == fileIdMap.end() )	       {
		  fileIdMap[ fileId ] = fileCounter++;
		  std::cout << fileId << std::endl;
	       }
	    }
	 }
      }

      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {      
	 collMap[i]->close();
      }
      return 0;

   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::Exception: " << exception.what() << std::endl;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception. " << std::endl;
      return 1;
   }
}
