/**
 * @file CollListPFN.cpp
 * @brief Utility to list the physical file names used by a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * $Id: CollListPFN.cpp,v 1.29 2009/09/17 15:12:25 mnowak Exp $
 */
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/ICollection.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "POOLCore/Exception.h"
#include "POOLCore/Token.h"

#include "CoralBase/MessageStream.h"

#include "Args2Container.h"
#include "SrcInfo.h"
#include "QueryInfo.h"
#include "CatalogInfo.h"

#include <iostream>
using namespace std;


int main(int argc, const char *argv[])
{
   string thisProgram("CollListPFN");

   // Convert argv to vector of strings
   vector<string> argv_v;
   for( int i=0; i<argc; ++ i )
      argv_v.push_back( argv[i] );

   try
   { 
      pool::CollectionService   collectionService;
      coral::MessageStream log( thisProgram );
 
      // vector of CmdLineArg objects
      Args2Container argsVec(thisProgram, true, &log);
      argsVec.desc << thisProgram
                   << " lists Physical File Names of POOL files referenced from the source event collection(s). "
                   << endl;

      QualList markers;
      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      argsVec.push_back(&cmdLineArgs); 

      CatalogInfo catinfo; argsVec.push_back(&catinfo);
      QueryInfo queryinfo; argsVec.push_back(&queryinfo);
      SrcInfo srcinfo;     argsVec.push_back(&srcinfo);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;

      // Fill appropriate vectors based on *specific* cmdLineArgs
      map < string, pair <int,int> >::iterator mit=cmdLineArgs.begin();
      while( mit != cmdLineArgs.end() ) {
         string qual = mit->first;
         //int ifirst = mit->second.first;
         log << coral::Error << "Unrecognized option: " << qual << endl
             << "Use -help for option list" << coral::MessageStream::endmsg;
         exit(-1); 
         ++mit;
      } 

      catinfo.setCatalogs( &collectionService );
      pool::IFileCatalog	FC;
      catinfo.setFileCatalogs( &FC );
      FC.start();

      std::map<int, pool::ICollection*> collMap;
      std::map< string, string > guidToPfn;
      std::map< std::string, int > pfnMap;
      std::map< std::string, int > notFoundGuids;
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {
         bool readOnly( true );
         pool::ICollection *collection = collectionService.handle( srcinfo.name(i), srcinfo.type(i), srcinfo.connect(), readOnly, 0 );
         collMap[i] = collection;

         log << coral::Debug
             << " Creating query for the collection" << coral::MessageStream::endmsg;
         pool::ICollectionQuery *collQuery = collection->newQuery();
         collQuery->setCondition( queryinfo.query() );
         if( queryinfo.queryOptions().size() ) {
            collQuery->addToOutputList( queryinfo.queryOptions() );
         } else {
            collQuery->selectAllTokens();
         }
         log << coral::Debug << " Executing the query " << coral::MessageStream::endmsg;
         pool::ICollectionCursor& cursor = collQuery->execute();

         while( cursor.next() ) {
            std::string dbFileId = cursor.eventRef().dbID();
            if( notFoundGuids.find( dbFileId ) != notFoundGuids.end() ) {
               // "known" unknown guid
               notFoundGuids[dbFileId]++;
            } else {
	       if( guidToPfn.find( dbFileId ) != guidToPfn.end() ) {
		  pfnMap[ guidToPfn[dbFileId] ] ++;
	       } else {
		  // new GUID
		  std::string bestPFN;
		  std::string fileType;
		  pool::IFCAction action;
		  FC.setAction( action );
		  action.lookupBestPFN( dbFileId, 
					pool::FileCatalog::READ, 
					pool::FileCatalog::SEQUENTIAL, 
					bestPFN, 
					fileType );
		  if( bestPFN.size() ) {
		     // found a name
		     guidToPfn[ dbFileId ] = bestPFN;
		     pfnMap[ bestPFN ] = 1;
		  } else {
		     notFoundGuids[dbFileId] = 1;
		  }
	       }
	    }
	 }
      }
      if( notFoundGuids.size() ) {
         log << coral::Warning << "Found " << notFoundGuids.size()
                   << " unresolved POOL file GUID(s)" << coral::MessageStream::endmsg;
	 std::map< std::string, int >::iterator mapIter;
 	 for( mapIter = notFoundGuids.begin(); mapIter != notFoundGuids.end(); ++mapIter ) {
	    log<< coral::Debug << "   unresolved GUID: " << mapIter->first << coral::MessageStream::endmsg;
	 }
      }
      if( !pfnMap.size() ) {
         log << coral::Warning << "No PFNs found - is your POOL file catalog correctly specified?" << coral::MessageStream::endmsg;
      } else {
         log << coral::Info << "Found PFNs referenced from the collection(s): " << coral::MessageStream::endmsg;
         std::map< std::string, int >::iterator mapIter;
         for( mapIter = pfnMap.begin(); mapIter != pfnMap.end(); ++mapIter ) {
            std::cout << mapIter->first << std::endl;
         }
      }
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {      
         collMap[i]->close();
      }
      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::Exception: " << exception.what() << std::endl;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception. " << std::endl;
      return 1;
   }
}
