/**
 * @file CollSplitByGUID.cpp
 * @brief Utility to list the file GUIDs used by a POOL collection and split the
 * collection into sub-collections by GUID
 * @author K. Karr <Kristo.Karr@cern.ch>, C. Nicholson <Caitriana.Nicholson@cern.ch>
 * $Id: CollSplitByGUID.cpp,v 1.9 2009/09/17 15:12:25 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionDataEditor.h"

#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"

#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ITransaction.h"

#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "Args2Container.h"
#include "SrcInfo.h"
#include "QueryInfo.h"
#include "CatalogInfo.h"
#include "MetaInfo.h"

#include <iostream>
#include <sstream>
#include <string>
#include <map>
#include <memory>   // for auto_ptr
#define AUTO_PTR  auto_ptr
 

using namespace std;
using namespace pool;


int main(int argc, const char *argv[])
{
   string thisProgram("CollSplitByGUID");

   // Convert argv to vector of strings
   vector<string> argv_v;
   for( int i=0; i<argc; ++ i )
      argv_v.push_back( argv[i] );

   try
   {
      bool noMetadata = false;
      // src collection info
      std::vector<int> srcCountVec;
      // dst collection info
      std::vector<bool> dstCollExistVec;
      // output modifiers
      long long minEvents = -1;
      //unsigned int numEventsPerCommit = 10000;
      unsigned int numEventsPerCommit = static_cast<unsigned int>(-1);
      int numRowsCached = 0;
      std::vector<std::string> inputQuery;
      map<string,string> srcMetadata;

      coral::MessageStream log( thisProgram );
      pool::CollectionService   collectionService;

      // vector of CmdLineArg objects
      Args2Container argsVec(thisProgram,true);
      argsVec.desc << thisProgram << " is a tool for querying an event collection, or "
                   << "collections, and storing the results in a number of output collections, "
                   << "one for each different event file GUID. " << endl
                   << "Currently, these sub-collections are output as LOCAL, "
                   << "ROOT-based collections with a fixed name and will appear in the directory from "
                   << "which the program was executed." << endl;
 
      // list of CollAppend *specific* cli keys and their argument properties
      QualList markers;
      markers.insert( make_pair("-minevents",    ArgQual()) );
      markers.insert( make_pair("-nevtpercommit",ArgQual()) );
      markers.insert( make_pair("-nevtcached",   ArgQual()) );
      markers.insert( make_pair("-nometadata",   ArgQual(0)) );
      markers.insert( make_pair("-splitref",     ArgQual()) );

      markers["-splitref"].desc << "Name of ref to use for boundaries of split (DEFAULT=primary ref)"; 
      markers["-minevents"].desc << "minimum number of events required to create a separate output collection for a particular GUID";
      markers["-nevtpercommit"].desc << "Max. number of events to process between "
		                     << "output transaction commits (default is infinity)";
      markers["-nevtcached"].desc << "size of the insert buffer for bulk operations "
		                  << "DEFAULT = 0 (no bulk operations)";
      markers["-nometadata"].desc << "Ignore source metadata. (will still accept new metadata from the command line -metadata option)";
 
      
      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      argsVec.push_back(&cmdLineArgs);  // Add it to the list

      // Classes with shared cli keys and their argument properties
      // Add them to the list
      CatalogInfo catinfo; argsVec.push_back(&catinfo);
      QueryInfo queryinfo; argsVec.push_back(&queryinfo);
      // DstInfo dstinfo; argsVec.push_back(&dstinfo);    // not supported yet
      SrcInfo srcinfo; argsVec.push_back(&srcinfo);
      MetaInfo metainfo; argsVec.push_back(&metainfo);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;

      std::string splitRef = "";
      // Fill appropriate vectors based on CollAppend *specific* cmdLineArgs
      map < string, pair <int,int> >::iterator mit=cmdLineArgs.begin();
      // bool had_copymode = false;
      while( mit != cmdLineArgs.end() ) {
         string qual = mit->first;
         int ifirst = mit->second.first;
         // cout << "**********  processing option: " << qual << " " << argv_v[ifirst] << endl;
         //int ilast  = mit->second.second;
         //cout << "Parsing " << qual << " " << ifirst << " " << ilast << endl;
         if (qual == "-minevents") {
            istringstream  str( argv_v[ifirst] );
	    str >> minEvents;
         }
         else if (qual == "-nometadata") {
	    noMetadata = true;
         }
         else if (qual == "-splitref") {
	    splitRef = argv_v[ifirst];
         }
         // ---
         // EXPERT OPTIONS
         // For tuning purposes the number of events between commits can be specified
         // ---
         else if (qual == "-nevtpercommit") {
	    numEventsPerCommit = atoi( argv[ifirst] );
         }
         else if (qual == "-nevtcached") {
	    numRowsCached = atoi( argv[ifirst] );
         }
         else {
            log << coral::Error << "Unrecognized option: " << qual << endl
                << "Use -help for option list" << coral::MessageStream::endmsg;
            exit(-1); 
         }

         ++mit;
      } 

      catinfo.setCatalogs( &collectionService );
      
      std::map<int, pool::ICollection*> collMap;
      int collectionNumber = 1;

      // for each input collection, loop through events and write out sub-collections
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ )
      {
       
         std::map< std::string, int > fileIdMap;
         std::map< std::string, pool::ICollection*> subCollMap;  // map to keep guid --> sub-collection
         std::multimap< std::string, std::string> invCollMap;  // map to keep sub-collection name --> guid

         log << coral::Info
             << "Opening source collection " << srcinfo.name(i)
             << " of type " << srcinfo.type(i) <<  coral::MessageStream::endmsg;
         bool readOnly( true );
         pool::ICollection *collection = collectionService.handle( srcinfo.name(i), srcinfo.type(i), srcinfo.connect(), readOnly );
         collMap[i] = collection;

         const pool::ICollectionDescription &firstDesc = collMap[0]->description();
         const pool::ICollectionDescription &nextDesc = collection->description();


         if( (queryinfo.query() != "" || queryinfo.queryOptions() != "") && !nextDesc.equals(firstDesc) ) {
            log << coral::Warning << " The schemas of one or more "
                << "input collections are different and a query has been "
                << "requested. This may lead to unexpected behavior."
                << coral::MessageStream::endmsg; 
         }

         pool::ICollectionQuery *collQuery = collection->newQuery();
         pool::ICollectionQuery *collQuery2 = collection->newQuery();
         collQuery->setCondition( queryinfo.query() );
         collQuery2->setCondition( queryinfo.query() );
         if( queryinfo.queryOptions().size() ) {
            collQuery->addToOutputList( queryinfo.queryOptions() );
            collQuery2->addToOutputList( queryinfo.queryOptions() );
         } else {
            collQuery->selectAll();
            collQuery2->selectAll();
         }
         
         log << coral::Info << "Executing query for the source collection" << std::endl; 
         pool::ICollectionCursor& cursor = collQuery->execute();

         // set parameters for the sub-collections. 
         // currently caters only for writing local, Root-based collections.
         std::string subCollType = "RootCollection";
         std::string subCollConnect = "";

         // first loop: for each event, find the GUID / fileId and 
         // count how many events match that GUID
         int totalEvents = 0;
         int numFiles = 0;

         // Token name to split on (if not specified, use default for each source coll)
         string refname = ( splitRef.size()? splitRef : collection->description().eventReferenceColumnName() );
         while( cursor.next() ) {
            const pool::TokenList &tokens = cursor.currentRow().tokenList();
            for( pool::TokenList::const_iterator iter = tokens.begin(); iter != tokens.end(); ++iter ) {
               std::string name = iter.tokenName();
               if( name == refname ) {   
                  std::string fileId = iter->dbID();
                  std::map< std::string, int >::iterator mapIter;
                  mapIter = fileIdMap.find( fileId );
                  if( mapIter == fileIdMap.end() )        {
                     fileIdMap[ fileId ] = numFiles++;
                  }
                  else {
                     fileIdMap[ fileId ] += 1;
                  }
               }
            }
            totalEvents++;
         }


         // make suitable collections, each with nEvents >= minEvents
         int heldEvents = 0;
         int currentFile = 0;
         bool lastCollection = false;
         std::vector<std::string> heldFileIds;   //to hold GUIDs until minEvents reached 
         pool::ICollection* previousCollection=0;

         std::map< std::string, int >::iterator mapIter = fileIdMap.begin();
         while ( mapIter != fileIdMap.end()) 
         {
            std::string fileId = mapIter->first;
            int thisNumEvents = mapIter->second;
            int counter = thisNumEvents + heldEvents;
            currentFile += 1;
            if ( currentFile == numFiles )
               lastCollection = true;

            // test for various conditions in which a new sub-collection ought to be created
            if( counter >= minEvents || lastCollection == true ) 
            {
               if( collectionNumber == 1 || !lastCollection || counter >= minEvents )
               {
                  //create sub-collection
                  std::stringstream collectionNameStream; 
                  collectionNameStream << "sub_collection_" << collectionNumber; 
                  std::string subCollName = collectionNameStream.str();
      
                  pool::CollectionDescription  destDescription( srcinfo.name(i), "type" );
                  destDescription = nextDesc;
                  pool::CollectionDescription newDestDesc( destDescription );
                  newDestDesc.setName( subCollName );
                  newDestDesc.setType( subCollType );               
                  newDestDesc.setConnection( subCollConnect );
                  pool::ICollection* subCollection = collectionService.create( newDestDesc, true );
                  subCollection->dataEditor().setRowCacheSize( 0 );                                                       

                  // map to appropriate GUID
                  subCollMap[ fileId ] = subCollection;
                  invCollMap.insert( std::pair<std::string, std::string>(subCollName, fileId));
                  for (unsigned int i=0; i<heldFileIds.size(); i++)
                  { 
                     std::string thisId = heldFileIds[i];
                     subCollMap[ thisId ] = subCollection; 
                     invCollMap.insert( std::pair<std::string, std::string>(subCollName, thisId));
                  }
                  heldEvents = 0;
                  heldFileIds.clear();
                  collectionNumber++;
                  previousCollection = subCollection;
               }
               else
               {
                  subCollMap[ fileId ] = previousCollection;
                  invCollMap.insert( std::pair<std::string, std::string>(previousCollection->description().name(), fileId));   
                  for (unsigned int i=0; i<heldFileIds.size(); i++)
                  {
                     std::string thisId = heldFileIds[i];
                     subCollMap[ thisId ] = previousCollection;
                     invCollMap.insert( std::pair<std::string, std::string>(previousCollection->description().name(), fileId));
                  }
               }
            }
            else
            {
               heldFileIds.push_back(fileId); 
               heldEvents += thisNumEvents;
            }
            ++mapIter;
         }

         std::string lastCollName = "";
         for( std::multimap<std::string, std::string>::const_iterator invCollIter = invCollMap.begin(); 
              invCollIter != invCollMap.end(); 
              ++invCollIter) 
         { 
            std::string thisCollName = invCollIter->first;
            if (thisCollName == lastCollName)
               log << coral::Info << " " << invCollIter->second;
            else {   
               if (invCollIter != invCollMap.begin())
                  log << coral::Info << coral::MessageStream::endmsg; 
               log << coral::Info << "Created new sub-collection " << thisCollName << " with files:\t" << invCollIter->second;
            }
            lastCollName = thisCollName;
         }
         log << coral::Info << coral::MessageStream::endmsg; 
 
         // second loop: write events to correct sub-collection
         pool::ICollectionCursor& cursor2 = collQuery2->execute();
         // Token name to split on (if not specified, use default for each source coll)
         const std::string srcRefName = ( splitRef.size()? splitRef : collection->description().eventReferenceColumnName() );
         while( cursor2.next() ) {
            // get file GUID from event
            const pool::TokenList &tokens = cursor2.currentRow().tokenList();
            for( pool::TokenList::const_iterator ti = tokens.begin(); ti != tokens.end(); ++ti ) {
               std::string name = ti.tokenName();
               if( name == srcRefName ) {
                  std::string guid = ti->dbID();
                  pool::CollectionRowBuffer &rowBuffer = subCollMap[guid]->dataEditor().rowBuffer();
                  rowBuffer.attributeList() = cursor2.currentRow().attributeList();
                  const std::string& dstRefName = subCollMap[guid]->description().eventReferenceColumnName();
                  if( ti.tokenName() == srcRefName ) {
                     ti->setData( &rowBuffer.tokenList()[ dstRefName ] );
                  } else {
                     ti->setData( &rowBuffer.tokenList()[ ti.tokenName() ] );
                  }
                  subCollMap[guid]->dataEditor().insertRow( rowBuffer );
               }
            }
         }

         log << coral::Info << "Finished writing all events from input collection " << collection->description().name() << coral::MessageStream::endmsg;
 

         // commit changes and close all sub-collections
         std::map< std::string, pool::ICollection* >::iterator closeIter = subCollMap.begin();
         while ( closeIter != subCollMap.end() )
         {
            std::string guid = closeIter->first;
            pool::ICollection* subCollection = closeIter->second;
            subCollection->commit();  
            subCollection->close();
            ++closeIter;
         }

      } // for each input collection

      for( unsigned int i=0; i< srcinfo.nSrc(); i++ ) {      
         collMap[i]->close();
      }
      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::Exception: " << exception.what() << std::endl;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception. " << std::endl;
      return 1;
   }
}
